import sys
from unittest import IsolatedAsyncioTestCase

import pytest

from ...askar.profile import AskarProfileSession
from ...ledger.base import BaseLedger
from ...ledger.error import LedgerError
from ...tests import mock
from ...utils.testing import create_test_profile
from .. import argparse
from .. import ledger as test_module

TEST_DID = "55GkHamhTU1ZbTbV2ab9DE"
TEST_GENESIS = "GENESIS"


class TestLedgerConfig(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.profile = await create_test_profile(
            settings={
                "wallet-type": "askar-anoncreds",
                "tails_server_base_url": "http://tails-server.com",
            }
        )

    async def test_fetch_genesis_transactions(self):
        with mock.patch.object(test_module, "fetch", mock.CoroutineMock()):
            await test_module.fetch_genesis_transactions("http://1.2.3.4:9000/genesis")

    async def test_fetch_genesis_transactions_x(self):
        with mock.patch.object(
            test_module, "fetch", mock.CoroutineMock(return_value=TEST_GENESIS)
        ) as mock_fetch:
            mock_fetch.side_effect = test_module.FetchError("404 Not Found")
            with self.assertRaises(test_module.ConfigError):
                await test_module.fetch_genesis_transactions(
                    "http://1.2.3.4:9000/genesis"
                )

    async def test_get_genesis_url(self):
        settings = {
            "ledger.genesis_url": "00000000000000000000000000000000",
        }
        with mock.patch.object(
            test_module,
            "fetch_genesis_transactions",
            mock.CoroutineMock(return_value=TEST_GENESIS),
        ):
            await test_module.get_genesis_transactions(settings)
        self.assertEqual(settings["ledger.genesis_transactions"], TEST_GENESIS)

    async def test_get_genesis_file(self):
        settings = {
            "ledger.genesis_file": "/tmp/genesis/path",
        }
        with mock.patch("builtins.open", mock.MagicMock()) as mock_open:
            mock_open.return_value = mock.MagicMock(
                __enter__=mock.MagicMock(
                    return_value=mock.MagicMock(
                        read=mock.MagicMock(return_value=TEST_GENESIS)
                    )
                )
            )
            await test_module.get_genesis_transactions(settings)
        self.assertEqual(settings["ledger.genesis_transactions"], TEST_GENESIS)

    async def test_get_genesis_file_io_x(self):
        settings = {
            "ledger.genesis_file": "/tmp/genesis/path",
        }

        with mock.patch("builtins.open", mock.MagicMock()) as mock_open:
            mock_open.side_effect = IOError("no read permission")
            with self.assertRaises(test_module.ConfigError):
                await test_module.get_genesis_transactions(settings)

    async def test_ledger_config_no_taa_accept(self):
        settings = {
            "ledger.genesis_transactions": TEST_GENESIS,
            "default_endpoint": "http://1.2.3.4:8051",
        }
        mock_ledger = mock.MagicMock(BaseLedger, autospec=True)
        mock_ledger.get_txn_author_agreement = mock.CoroutineMock(
            return_value={
                "taa_required": True,
                "taa_record": {"digest": b"ffffffffffffffffffffffffffffffffffffffff"},
            }
        )
        mock_ledger.get_latest_txn_author_acceptance = mock.CoroutineMock(
            return_value={"digest": b"1234567890123456789012345678901234567890"}
        )

        mock_ledger.read_only = False

        self.profile = await create_test_profile(settings=settings)

        self.profile.context.injector.bind_instance(BaseLedger, mock_ledger)

        with mock.patch.object(
            test_module, "accept_taa", mock.CoroutineMock()
        ) as mock_accept_taa:
            mock_accept_taa.return_value = False
            assert not await test_module.ledger_config(
                self.profile, TEST_DID, provision=True
            )

    @mock.patch.object(
        AskarProfileSession,
        "inject",
        mock.MagicMock(
            return_value=mock.MagicMock(set_did_endpoint=mock.CoroutineMock())
        ),
    )
    async def test_accept_taa(self):
        settings = {
            "ledger.genesis_transactions": TEST_GENESIS,
        }
        mock_ledger = mock.MagicMock(BaseLedger, autospec=True)
        mock_ledger.get_txn_author_agreement = mock.CoroutineMock(
            return_value={
                "taa_required": True,
                "taa_record": {"digest": b"ffffffffffffffffffffffffffffffffffffffff"},
            }
        )
        mock_ledger.get_latest_txn_author_acceptance = mock.CoroutineMock(
            return_value={"digest": b"1234567890123456789012345678901234567890"}
        )
        mock_ledger.update_endpoint_for_did = mock.CoroutineMock()
        mock_ledger.read_only = False

        self.profile = await create_test_profile(settings=settings)
        self.profile.context.injector.bind_instance(BaseLedger, mock_ledger)

        with mock.patch.object(
            test_module, "accept_taa", mock.CoroutineMock()
        ) as mock_accept_taa:
            mock_accept_taa.return_value = True
            await test_module.ledger_config(self.profile, TEST_DID, provision=True)
            mock_accept_taa.assert_awaited_once()

    @mock.patch.object(
        AskarProfileSession,
        "inject",
        mock.MagicMock(
            return_value=mock.MagicMock(
                set_did_endpoint=mock.CoroutineMock(
                    side_effect=LedgerError(
                        "Error cannot update endpoint when ledger is in read only mode"
                    )
                )
            )
        ),
    )
    async def test_ledger_config_read_only_skip_taa_accept(self):
        settings = {
            "ledger.genesis_transactions": TEST_GENESIS,
        }

        mock_ledger = mock.MagicMock(BaseLedger, autospec=True)
        mock_ledger.get_txn_author_agreement = mock.CoroutineMock()
        mock_ledger.get_latest_txn_author_acceptance = mock.CoroutineMock()
        mock_ledger.read_only = True

        self.profile = await create_test_profile(settings=settings)
        self.profile.context.injector.bind_instance(BaseLedger, mock_ledger)

        with mock.patch.object(test_module, "accept_taa", mock.CoroutineMock()):
            with self.assertRaises(test_module.ConfigError) as x_context:
                await test_module.ledger_config(self.profile, TEST_DID, provision=True)
            assert "ledger is in read only mode" in str(x_context.exception)
            mock_ledger.get_txn_author_agreement.assert_not_called()
            mock_ledger.get_latest_txn_author_acceptance.assert_not_called()

    @mock.patch.object(
        AskarProfileSession,
        "inject",
        mock.MagicMock(
            return_value=mock.MagicMock(set_did_endpoint=mock.CoroutineMock())
        ),
    )
    async def test_ledger_config_read_only_skip_profile_endpoint_publish(self):
        settings = {
            "ledger.genesis_url": "00000000000000000000000000000000",
            "profile_endpoint": "http://agent.ca",
        }

        mock_ledger = mock.MagicMock(BaseLedger, autospec=True)
        mock_ledger.get_txn_author_agreement = mock.CoroutineMock()
        mock_ledger.get_latest_txn_author_acceptance = mock.CoroutineMock()
        mock_ledger.read_only = True

        self.profile = await create_test_profile(settings=settings)
        self.profile.context.injector.bind_instance(BaseLedger, mock_ledger)

        with mock.patch.object(test_module, "accept_taa", mock.CoroutineMock()):
            await test_module.ledger_config(self.profile, TEST_DID, provision=True)
            mock_ledger.get_txn_author_agreement.assert_not_called()
            mock_ledger.get_latest_txn_author_acceptance.assert_not_called()
            mock_ledger.update_endpoint_for_did.assert_not_called()

    @mock.patch.object(
        AskarProfileSession,
        "inject",
        mock.MagicMock(
            return_value=mock.MagicMock(set_did_endpoint=mock.CoroutineMock())
        ),
    )
    async def test_ledger_config_read_write_skip_taa_endpoint_publish(self):
        settings = {
            "ledger.genesis_url": "00000000000000000000000000000000",
            "default_endpoint": "http://agent-default.ca",
            "profile_endpoint": "http://agent-profile.ca",
        }
        mock_ledger = mock.MagicMock(BaseLedger, autospec=True)
        mock_ledger.get_txn_author_agreement = mock.CoroutineMock(
            return_value={"taa_required": False}
        )
        mock_ledger.get_latest_txn_author_acceptance = mock.CoroutineMock()
        mock_ledger.read_only = False

        self.profile = await create_test_profile(settings=settings)
        self.profile.context.injector.bind_instance(BaseLedger, mock_ledger)

        with mock.patch.object(test_module, "accept_taa", mock.CoroutineMock()):
            await test_module.ledger_config(
                self.profile,
                public_did=TEST_DID,
                provision=False,
            )
            mock_ledger.get_txn_author_agreement.assert_called_once_with()
            mock_ledger.get_latest_txn_author_acceptance.assert_not_called()
            mock_ledger.update_endpoint_for_did.assert_called_once_with(
                TEST_DID,
                settings["profile_endpoint"],
                test_module.EndpointType.PROFILE,
            )

    async def test_load_multiple_genesis_transactions_from_config_a(self):
        TEST_GENESIS_TXNS = {
            "reqSignature": {},
            "txn": {
                "data": {
                    "data": {
                        "alias": "Node1",
                        "blskey": "4N8aUNHSgjQVgkpm8nhNEfDf6txHznoYREg9kirmJrkivgL4oSEimFF6nsQ6M41QvhM2Z33nves5vfSn9n1UwNFJBYtWVnHYMATn76vLuL3zU88KyeAYcHfsih3He6UHcXDxcaecHVz6jhCYz1P2UZn2bDVruL5wXpehgBfBaLKm3Ba",
                        "blskey_pop": "RahHYiCvoNCtPTrVtP7nMC5eTYrsUA8WjXbdhNc8debh1agE9bGiJxWBXYNFbnJXoXhWFMvyqhqhRoq737YQemH5ik9oL7R4NTTCz2LEZhkgLJzB3QRQqJyBNyv7acbdHrAT8nQ9UkLbaVL9NBpnWXBTw4LEMePaSHEw66RzPNdAX1",
                        "client_ip": "192.168.65.3",
                        "client_port": 9702,
                        "node_ip": "192.168.65.3",
                        "node_port": 9701,
                        "services": ["VALIDATOR"],
                    },
                    "dest": "Gw6pDLhcBcoQesN72qfotTgFa7cbuqZpkX3Xo6pLhPhv",
                },
                "metadata": {"from": "Th7MpTaRZVRYnPiabds81Y"},
                "type": "0",
            },
            "txnMetadata": {
                "seqNo": 1,
                "txnId": "fea82e10e894419fe2bea7d96296a6d46f50f93f9eeda954ec461b2ed2950b62",
            },
            "ver": "1",
        }
        TEST_MULTIPLE_LEDGER_CONFIG_LIST = [
            {
                "id": "sovrinMain",
                "is_production": True,
                "genesis_transactions": TEST_GENESIS_TXNS,
                "is_write": True,
                "keepalive": 5,
                "read_only": False,
                "socks_proxy": None,
                "pool_name": "sovrinMain",
                "endorser_did": "9QPa6tHvBHttLg6U4xvviv",
                "endorser_alias": "endorser_main",
            },
            {
                "id": "sovrinStaging",
                "is_production": True,
                "genesis_transactions": TEST_GENESIS_TXNS,
                "is_write": False,
                "keepalive": 5,
                "read_only": False,
                "socks_proxy": None,
                "pool_name": "sovrinStaging",
            },
            {
                "id": "sovrinTest",
                "is_production": True,
                "genesis_transactions": TEST_GENESIS_TXNS,
                "is_write": False,
                "keepalive": 5,
                "read_only": False,
                "socks_proxy": None,
                "pool_name": "sovrinTest",
            },
        ]
        settings = {
            "ledger.ledger_config_list": [
                {
                    "id": "sovrinMain",
                    "is_production": True,
                    "is_write": True,
                    "genesis_transactions": TEST_GENESIS_TXNS,
                    "endorser_did": "9QPa6tHvBHttLg6U4xvviv",
                    "endorser_alias": "endorser_main",
                },
                {
                    "id": "sovrinStaging",
                    "is_production": True,
                    "genesis_file": "/home/indy/ledger/sandbox/pool_transactions_genesis",
                },
                {
                    "id": "sovrinTest",
                    "is_production": True,
                    "genesis_url": "http://localhost:9000/genesis",
                },
            ],
        }
        with (
            mock.patch.object(
                test_module,
                "fetch_genesis_transactions",
                mock.CoroutineMock(return_value=TEST_GENESIS_TXNS),
            ),
            mock.patch("builtins.open", mock.MagicMock()) as mock_open,
        ):
            mock_open.return_value = mock.MagicMock(
                __enter__=mock.MagicMock(
                    return_value=mock.MagicMock(
                        read=mock.MagicMock(return_value=TEST_GENESIS_TXNS)
                    )
                )
            )
            await test_module.load_multiple_genesis_transactions_from_config(settings)
        self.assertEqual(
            settings["ledger.ledger_config_list"], TEST_MULTIPLE_LEDGER_CONFIG_LIST
        )

    async def test_load_multiple_genesis_transactions_from_config_b(self):
        TEST_GENESIS_TXNS = {
            "reqSignature": {},
            "txn": {
                "data": {
                    "data": {
                        "alias": "Node1",
                        "blskey": "4N8aUNHSgjQVgkpm8nhNEfDf6txHznoYREg9kirmJrkivgL4oSEimFF6nsQ6M41QvhM2Z33nves5vfSn9n1UwNFJBYtWVnHYMATn76vLuL3zU88KyeAYcHfsih3He6UHcXDxcaecHVz6jhCYz1P2UZn2bDVruL5wXpehgBfBaLKm3Ba",
                        "blskey_pop": "RahHYiCvoNCtPTrVtP7nMC5eTYrsUA8WjXbdhNc8debh1agE9bGiJxWBXYNFbnJXoXhWFMvyqhqhRoq737YQemH5ik9oL7R4NTTCz2LEZhkgLJzB3QRQqJyBNyv7acbdHrAT8nQ9UkLbaVL9NBpnWXBTw4LEMePaSHEw66RzPNdAX1",
                        "client_ip": "192.168.65.3",
                        "client_port": 9702,
                        "node_ip": "192.168.65.3",
                        "node_port": 9701,
                        "services": ["VALIDATOR"],
                    },
                    "dest": "Gw6pDLhcBcoQesN72qfotTgFa7cbuqZpkX3Xo6pLhPhv",
                },
                "metadata": {"from": "Th7MpTaRZVRYnPiabds81Y"},
                "type": "0",
            },
            "txnMetadata": {
                "seqNo": 1,
                "txnId": "fea82e10e894419fe2bea7d96296a6d46f50f93f9eeda954ec461b2ed2950b62",
            },
            "ver": "1",
        }
        TEST_MULTIPLE_LEDGER_CONFIG_LIST = [
            {
                "id": "sovrinMain",
                "is_production": True,
                "genesis_transactions": TEST_GENESIS_TXNS,
                "is_write": False,
                "keepalive": 5,
                "read_only": False,
                "socks_proxy": None,
                "pool_name": "sovrinMain",
            },
            {
                "id": "sovrinStaging",
                "is_production": True,
                "genesis_transactions": TEST_GENESIS_TXNS,
                "is_write": False,
                "keepalive": 5,
                "read_only": False,
                "socks_proxy": None,
                "pool_name": "sovrinStaging",
            },
            {
                "id": "sovrinTest",
                "is_production": True,
                "genesis_transactions": TEST_GENESIS_TXNS,
                "is_write": False,
                "keepalive": 5,
                "read_only": False,
                "socks_proxy": None,
                "pool_name": "sovrinTest",
            },
        ]
        settings = {
            "ledger.ledger_config_list": [
                {
                    "id": "sovrinMain",
                    "is_production": True,
                    "genesis_transactions": TEST_GENESIS_TXNS,
                },
                {
                    "id": "sovrinStaging",
                    "is_production": True,
                    "genesis_file": "/home/indy/ledger/sandbox/pool_transactions_genesis",
                },
                {
                    "id": "sovrinTest",
                    "is_production": True,
                    "genesis_url": "http://localhost:9001/genesis",
                },
            ],
            "ledger.genesis_url": "http://localhost:9000/genesis",
        }
        with (
            mock.patch.object(
                test_module,
                "fetch_genesis_transactions",
                mock.CoroutineMock(return_value=TEST_GENESIS_TXNS),
            ),
            mock.patch("builtins.open", mock.MagicMock()) as mock_open,
        ):
            mock_open.return_value = mock.MagicMock(
                __enter__=mock.MagicMock(
                    return_value=mock.MagicMock(
                        read=mock.MagicMock(return_value=TEST_GENESIS_TXNS)
                    )
                )
            )
            await test_module.load_multiple_genesis_transactions_from_config(settings)
        self.assertEqual(
            settings["ledger.ledger_config_list"], TEST_MULTIPLE_LEDGER_CONFIG_LIST
        )

    async def test_load_multiple_genesis_transactions_config_error_a(self):
        TEST_GENESIS_TXNS = {
            "reqSignature": {},
            "txn": {
                "data": {
                    "data": {
                        "alias": "Node1",
                        "blskey": "4N8aUNHSgjQVgkpm8nhNEfDf6txHznoYREg9kirmJrkivgL4oSEimFF6nsQ6M41QvhM2Z33nves5vfSn9n1UwNFJBYtWVnHYMATn76vLuL3zU88KyeAYcHfsih3He6UHcXDxcaecHVz6jhCYz1P2UZn2bDVruL5wXpehgBfBaLKm3Ba",
                        "blskey_pop": "RahHYiCvoNCtPTrVtP7nMC5eTYrsUA8WjXbdhNc8debh1agE9bGiJxWBXYNFbnJXoXhWFMvyqhqhRoq737YQemH5ik9oL7R4NTTCz2LEZhkgLJzB3QRQqJyBNyv7acbdHrAT8nQ9UkLbaVL9NBpnWXBTw4LEMePaSHEw66RzPNdAX1",
                        "client_ip": "192.168.65.3",
                        "client_port": 9702,
                        "node_ip": "192.168.65.3",
                        "node_port": 9701,
                        "services": ["VALIDATOR"],
                    },
                    "dest": "Gw6pDLhcBcoQesN72qfotTgFa7cbuqZpkX3Xo6pLhPhv",
                },
                "metadata": {"from": "Th7MpTaRZVRYnPiabds81Y"},
                "type": "0",
            },
            "txnMetadata": {
                "seqNo": 1,
                "txnId": "fea82e10e894419fe2bea7d96296a6d46f50f93f9eeda954ec461b2ed2950b62",
            },
            "ver": "1",
        }
        settings = {
            "ledger.ledger_config_list": [
                {
                    "id": "sovrinMain",
                    "is_production": True,
                    "genesis_transactions": TEST_GENESIS_TXNS,
                },
                {
                    "id": "sovrinStaging",
                    "is_production": True,
                    "genesis_file": "/home/indy/ledger/sandbox/pool_transactions_genesis",
                },
                {
                    "id": "sovrinTest",
                    "is_production": True,
                    "genesis_url": "http://localhost:9001/genesis",
                },
            ],
        }
        with (
            mock.patch.object(
                test_module,
                "fetch_genesis_transactions",
                mock.CoroutineMock(return_value=TEST_GENESIS_TXNS),
            ),
            mock.patch("builtins.open", mock.MagicMock()) as mock_open,
        ):
            mock_open.return_value = mock.MagicMock(
                __enter__=mock.MagicMock(
                    return_value=mock.MagicMock(
                        read=mock.MagicMock(return_value=TEST_GENESIS_TXNS)
                    )
                )
            )
            with self.assertRaises(test_module.ConfigError) as cm:
                await test_module.load_multiple_genesis_transactions_from_config(settings)
            assert "No writable ledger configured" in str(cm.exception)

    async def test_load_multiple_genesis_transactions_multiple_write(self):
        TEST_GENESIS_TXNS = {
            "reqSignature": {},
            "txn": {
                "data": {
                    "data": {
                        "alias": "Node1",
                        "blskey": "4N8aUNHSgjQVgkpm8nhNEfDf6txHznoYREg9kirmJrkivgL4oSEimFF6nsQ6M41QvhM2Z33nves5vfSn9n1UwNFJBYtWVnHYMATn76vLuL3zU88KyeAYcHfsih3He6UHcXDxcaecHVz6jhCYz1P2UZn2bDVruL5wXpehgBfBaLKm3Ba",
                        "blskey_pop": "RahHYiCvoNCtPTrVtP7nMC5eTYrsUA8WjXbdhNc8debh1agE9bGiJxWBXYNFbnJXoXhWFMvyqhqhRoq737YQemH5ik9oL7R4NTTCz2LEZhkgLJzB3QRQqJyBNyv7acbdHrAT8nQ9UkLbaVL9NBpnWXBTw4LEMePaSHEw66RzPNdAX1",
                        "client_ip": "192.168.65.3",
                        "client_port": 9702,
                        "node_ip": "192.168.65.3",
                        "node_port": 9701,
                        "services": ["VALIDATOR"],
                    },
                    "dest": "Gw6pDLhcBcoQesN72qfotTgFa7cbuqZpkX3Xo6pLhPhv",
                },
                "metadata": {"from": "Th7MpTaRZVRYnPiabds81Y"},
                "type": "0",
            },
            "txnMetadata": {
                "seqNo": 1,
                "txnId": "fea82e10e894419fe2bea7d96296a6d46f50f93f9eeda954ec461b2ed2950b62",
            },
            "ver": "1",
        }
        settings = {
            "ledger.ledger_config_list": [
                {
                    "id": "sovrinMain",
                    "is_production": True,
                    "is_write": True,
                    "genesis_transactions": TEST_GENESIS_TXNS,
                },
                {
                    "id": "sovrinStaging",
                    "is_production": True,
                    "is_write": True,
                    "genesis_file": "/home/indy/ledger/sandbox/pool_transactions_genesis",
                },
                {
                    "id": "sovrinTest",
                    "is_production": True,
                    "genesis_url": "http://localhost:9001/genesis",
                },
            ]
        }
        with (
            mock.patch.object(
                test_module,
                "fetch_genesis_transactions",
                mock.CoroutineMock(return_value=TEST_GENESIS_TXNS),
            ),
            mock.patch("builtins.open", mock.MagicMock()) as mock_open,
        ):
            mock_open.return_value = mock.MagicMock(
                __enter__=mock.MagicMock(
                    return_value=mock.MagicMock(
                        read=mock.MagicMock(return_value=TEST_GENESIS_TXNS)
                    )
                )
            )
            await test_module.load_multiple_genesis_transactions_from_config(settings)

    async def test_load_multiple_genesis_transactions_from_config_io_x(self):
        TEST_GENESIS_TXNS = {
            "reqSignature": {},
            "txn": {
                "data": {
                    "data": {
                        "alias": "Node1",
                        "blskey": "4N8aUNHSgjQVgkpm8nhNEfDf6txHznoYREg9kirmJrkivgL4oSEimFF6nsQ6M41QvhM2Z33nves5vfSn9n1UwNFJBYtWVnHYMATn76vLuL3zU88KyeAYcHfsih3He6UHcXDxcaecHVz6jhCYz1P2UZn2bDVruL5wXpehgBfBaLKm3Ba",
                        "blskey_pop": "RahHYiCvoNCtPTrVtP7nMC5eTYrsUA8WjXbdhNc8debh1agE9bGiJxWBXYNFbnJXoXhWFMvyqhqhRoq737YQemH5ik9oL7R4NTTCz2LEZhkgLJzB3QRQqJyBNyv7acbdHrAT8nQ9UkLbaVL9NBpnWXBTw4LEMePaSHEw66RzPNdAX1",
                        "client_ip": "192.168.65.3",
                        "client_port": 9702,
                        "node_ip": "192.168.65.3",
                        "node_port": 9701,
                        "services": ["VALIDATOR"],
                    },
                    "dest": "Gw6pDLhcBcoQesN72qfotTgFa7cbuqZpkX3Xo6pLhPhv",
                },
                "metadata": {"from": "Th7MpTaRZVRYnPiabds81Y"},
                "type": "0",
            },
            "txnMetadata": {
                "seqNo": 1,
                "txnId": "fea82e10e894419fe2bea7d96296a6d46f50f93f9eeda954ec461b2ed2950b62",
            },
            "ver": "1",
        }
        settings = {
            "ledger.ledger_config_list": [
                {
                    "id": "sovrinMain",
                    "is_production": True,
                    "genesis_transactions": TEST_GENESIS_TXNS,
                },
                {
                    "id": "sovrinStaging",
                    "is_production": True,
                    "genesis_file": "/home/indy/ledger/sandbox/pool_transactions_genesis",
                },
                {
                    "id": "sovrinTest",
                    "is_production": True,
                    "genesis_url": "http://localhost:9000/genesis",
                },
            ],
        }
        with (
            mock.patch.object(
                test_module,
                "fetch_genesis_transactions",
                mock.CoroutineMock(return_value=TEST_GENESIS_TXNS),
            ),
            mock.patch("builtins.open", mock.MagicMock()) as mock_open,
        ):
            mock_open.side_effect = IOError("no read permission")
            with self.assertRaises(test_module.ConfigError):
                await test_module.load_multiple_genesis_transactions_from_config(settings)

    @mock.patch("sys.stdout")
    async def test_ledger_accept_taa_not_tty_not_accept_config(self, mock_stdout):
        mock_stdout.isatty = mock.MagicMock(return_value=False)
        self.profile = await create_test_profile()

        taa_info = {
            "taa_record": {"version": "1.0", "text": "Agreement"},
            "aml_record": {"aml": ["wallet_agreement", "on_file"]},
        }

        assert not await test_module.accept_taa(
            None, self.profile, taa_info, provision=False
        )

    async def test_ledger_accept_taa_tty(self):
        sys.stdout.isatty = mock.MagicMock(return_value=True)
        self.profile = await create_test_profile()

        taa_info = {
            "taa_record": {"version": "1.0", "text": "Agreement"},
            "aml_record": {"aml": ["wallet_agreement", "on_file"]},
        }

        with (
            mock.patch.object(
                test_module.prompt_toolkit, "prompt", mock.CoroutineMock()
            ) as mock_prompt,
        ):
            mock_prompt.side_effect = EOFError()
            assert not await test_module.accept_taa(
                None, self.profile, taa_info, provision=False
            )

        with (
            mock.patch.object(
                test_module.prompt_toolkit, "prompt", mock.CoroutineMock()
            ) as mock_prompt,
        ):
            mock_prompt.return_value = "x"
            assert not await test_module.accept_taa(
                None, self.profile, taa_info, provision=False
            )

        with (
            mock.patch.object(
                test_module.prompt_toolkit, "prompt", mock.CoroutineMock()
            ) as mock_prompt,
        ):
            mock_ledger = mock.MagicMock(accept_txn_author_agreement=mock.CoroutineMock())
            mock_prompt.return_value = ""
            assert await test_module.accept_taa(
                mock_ledger, self.profile, taa_info, provision=False
            )

    async def test_ledger_accept_taa(self):
        taa_info = {
            "taa_record": {"version": "1.0", "text": "Agreement"},
            "aml_record": {"aml": {"wallet_agreement": "", "on_file": ""}},
        }

        # Incorrect version
        self.profile = await create_test_profile(
            {
                "ledger.taa_acceptance_mechanism": "wallet_agreement",
                "ledger.taa_acceptance_version": "1.5",
            }
        )
        with pytest.raises(LedgerError):
            await test_module.accept_taa(None, self.profile, taa_info, provision=False)

        # Incorrect mechanism
        self.profile = await create_test_profile(
            {
                "ledger.taa_acceptance_mechanism": "not_exist",
                "ledger.taa_acceptance_version": "1.0",
            }
        )
        with pytest.raises(LedgerError):
            await test_module.accept_taa(None, self.profile, taa_info, provision=False)

        # Valid
        self.profile = await create_test_profile(
            {
                "ledger.taa_acceptance_mechanism": "on_file",
                "ledger.taa_acceptance_version": "1.0",
            }
        )
        mock_ledger = mock.MagicMock(BaseLedger, autospec=True)
        mock_ledger.get_txn_author_agreement = mock.CoroutineMock()
        assert await test_module.accept_taa(
            mock_ledger, self.profile, taa_info, provision=False
        )

    async def test_ledger_config(self):
        """Test required argument parsing."""

        parser = argparse.create_argument_parser()
        group = argparse.LedgerGroup()
        group.add_arguments(parser)

        with mock.patch.object(parser, "exit") as exit_parser:
            parser.parse_args(["-h"])
            exit_parser.assert_called_once()

        result = parser.parse_args(
            [
                "--genesis-url",
                "http://1.2.3.4:9000/genesis",
                "--ledger-keepalive",
                "10",
            ]
        )

        assert result.ledger_keepalive == 10
        assert result.genesis_url == "http://1.2.3.4:9000/genesis"

        settings = group.get_settings(result)

        assert settings["ledger.keepalive"] == result.ledger_keepalive
        assert settings["ledger.genesis_url"] == result.genesis_url



"""Test json-ld credential."""

import json
from unittest import IsolatedAsyncioTestCase

from ....utils.testing import create_test_profile
from ....wallet.base import BaseWallet
from ....wallet.key_type import ED25519
from .. import credential as test_module
from ..create_verify_data import DroppedAttributeError
from ..credential import did_key, sign_credential, verify_credential
from ..error import BadJWSHeaderError, SignatureTypeError
from . import (
    TEST_SEED,
    TEST_SIGN_ERROR_OBJS,
    TEST_SIGN_OBJS,
    TEST_VERIFY_ERROR,
    TEST_VERIFY_OBJS,
    TEST_VERKEY,
)


class TestCredential(IsolatedAsyncioTestCase):
    async def test_did_key(self):
        did_key = test_module.did_key(TEST_VERKEY)
        assert did_key.startswith("did:key:z")
        assert did_key == test_module.did_key(did_key)

    async def test_verify_jws_header(self):
        test_module.verify_jws_header(
            json.loads(
                test_module.b64decode(
                    "eyJhbGciOiJFZERTQSIsImI2NCI6ZmFsc2UsImNyaXQiOlsiYjY0Il19"
                )
            )
        )

        with self.assertRaises(test_module.BadJWSHeaderError):
            test_module.verify_jws_header(
                json.loads(
                    test_module.b64decode(  # {... "b64": True ...}
                        "eyJhbGciOiJFZERTQSIsImI2NCI6dHJ1ZSwiY3JpdCI6WyJiNjQiXX0="
                    )
                )
            )


class TestOps(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        async with self.profile.session() as session:
            self.wallet = session.inject(BaseWallet)
            await self.wallet.create_signing_key(ED25519, TEST_SEED)

    async def test_verify_credential(self):
        async with self.profile.session() as session:
            for input_ in TEST_VERIFY_OBJS:
                assert await verify_credential(
                    session,
                    input_.get("doc"),
                    input_.get("verkey"),
                )

    async def test_sign_credential(self):
        async with self.profile.session() as session:
            for input_ in TEST_SIGN_OBJS:
                result = await sign_credential(
                    session,
                    input_.get("doc"),
                    input_.get("options"),
                    TEST_VERKEY,
                )
                assert "proof" in result.keys()
                assert "jws" in result.get("proof", {}).keys()

    async def test_sign_dropped_attribute_exception(self):
        async with self.profile.session() as session:
            for input_ in TEST_SIGN_ERROR_OBJS:
                with self.assertRaises(DroppedAttributeError) as context:
                    await sign_credential(
                        session,
                        input_.get("doc"),
                        input_.get("options"),
                        TEST_VERKEY,
                    )
                assert "attribute2drop" in str(context.exception)

    async def test_signature_option_type(self):
        async with self.profile.session() as session:
            for input_ in TEST_SIGN_OBJS:
                with self.assertRaises(SignatureTypeError):
                    input_["options"]["type"] = "Ed25519Signature2038"
                    await sign_credential(
                        session,
                        input_.get("doc"),
                        input_.get("options"),
                        TEST_VERKEY,
                    )

    async def test_invalid_jws_header(self):
        with self.assertRaises(BadJWSHeaderError):
            async with self.profile.session() as session:
                await verify_credential(
                    session,
                    TEST_VERIFY_ERROR.get("doc"),
                    TEST_VERIFY_ERROR.get("verkey"),
                )

    async def test_did_key(self):
        for verkey in (
            "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx",
            "did:key:z3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx",
        ):
            assert did_key(verkey).startswith("did:key:z")



"""Out-of-band handling admin routes."""

import json
import logging

from aiohttp import web
from aiohttp_apispec import (
    docs,
    match_info_schema,
    querystring_schema,
    request_schema,
    response_schema,
)
from marshmallow import fields, validate
from marshmallow.exceptions import ValidationError

from ....admin.decorators.auth import tenant_authentication
from ....admin.request_context import AdminRequestContext
from ....messaging.models.base import BaseModelError
from ....messaging.models.openapi import OpenAPISchema
from ....messaging.valid import UUID4_EXAMPLE, UUID4_VALIDATE
from ....storage.error import StorageError, StorageNotFoundError
from ...didcomm_prefix import DIDCommPrefix
from ...didexchange.v1_0.manager import DIDXManager, DIDXManagerError
from .manager import OutOfBandManager, OutOfBandManagerError
from .message_types import SPEC_URI
from .messages.invitation import HSProto, InvitationMessage, InvitationMessageSchema
from .models.invitation import InvitationRecordSchema
from .models.oob_record import OobRecordSchema

LOGGER = logging.getLogger(__name__)


class OutOfBandModuleResponseSchema(OpenAPISchema):
    """Response schema for Out of Band Module."""


class OobIdQueryStringSchema(OpenAPISchema):
    """Parameters and validators for fetch invitation request query string."""

    oob_id = fields.Str(
        required=True,
        validate=UUID4_VALIDATE,
        metadata={"description": "The Out of Band id to fetch"},
    )


class InvitationCreateQueryStringSchema(OpenAPISchema):
    """Parameters and validators for create invitation request query string."""

    auto_accept = fields.Boolean(
        required=False,
        metadata={"description": "Auto-accept connection (defaults to configuration)"},
    )
    multi_use = fields.Boolean(
        required=False,
        metadata={"description": "Create invitation for multiple use (default false)"},
    )
    create_unique_did = fields.Boolean(
        required=False,
        metadata={"description": "Create unique DID for this invitation (default false)"},
    )


class InvitationCreateRequestSchema(OpenAPISchema):
    """Invitation create request Schema."""

    class AttachmentDefSchema(OpenAPISchema):
        """Attachment Schema."""

        _id = fields.Str(
            data_key="id",
            metadata={
                "description": "Attachment identifier",
                "example": "attachment-0",
            },
        )
        _type = fields.Str(
            data_key="type",
            validate=validate.OneOf(["credential-offer", "present-proof"]),
            metadata={"description": "Attachment type", "example": "present-proof"},
        )

    attachments = fields.Nested(
        AttachmentDefSchema,
        many=True,
        required=False,
        metadata={"description": "Optional invitation attachments"},
    )
    handshake_protocols = fields.List(
        fields.Str(
            validate=lambda hsp: HSProto.get(hsp) is not None,
            metadata={
                "description": "Handshake protocol to specify in invitation",
                "example": DIDCommPrefix.qualify_current(HSProto.RFC23.name),
            },
        ),
        required=False,
    )
    accept = fields.List(
        fields.Str(),
        required=False,
        metadata={
            "description": (
                "List of mime type in order of preference that should be use in"
                " responding to the message"
            ),
            "example": ["didcomm/aip1", "didcomm/aip2;env=rfc19"],
        },
    )
    use_public_did = fields.Boolean(
        dump_default=False,
        metadata={
            "description": "Whether to use public DID in invitation",
            "example": False,
        },
    )
    use_did = fields.Str(
        required=False,
        metadata={
            "description": "DID to use in invitation",
            "example": "did:example:123",
        },
    )
    use_did_method = fields.Str(
        required=False,
        validate=validate.OneOf(DIDXManager.SUPPORTED_USE_DID_METHODS),
        metadata={
            "description": "DID method to use in invitation",
            "example": "did:peer:2",
        },
    )
    metadata = fields.Dict(
        required=False,
        metadata={
            "description": (
                "Optional metadata to attach to the connection created with the"
                " invitation"
            )
        },
    )
    my_label = fields.Str(
        required=False,
        metadata={
            "description": "Label for connection invitation",
            "example": "Invitation to Barry",
        },
    )
    protocol_version = fields.Str(
        required=False,
        metadata={"description": "OOB protocol version", "example": "1.1"},
    )
    alias = fields.Str(
        required=False,
        metadata={"description": "Alias for connection", "example": "Barry"},
    )
    mediation_id = fields.Str(
        required=False,
        validate=UUID4_VALIDATE,
        metadata={
            "description": "Identifier for active mediation record to be used",
            "example": UUID4_EXAMPLE,
        },
    )
    goal_code = fields.Str(
        required=False,
        metadata={
            "description": (
                "A self-attested code the receiver may want to display to the user or"
                " use in automatically deciding what to do with the out-of-band message"
            ),
            "example": "issue-vc",
        },
    )
    goal = fields.Str(
        required=False,
        metadata={
            "description": (
                "A self-attested string that the receiver may want to display to the"
                " user about the context-specific goal of the out-of-band message"
            ),
            "example": "To issue a Faber College Graduate credential",
        },
    )


class InvitationReceiveQueryStringSchema(OpenAPISchema):
    """Parameters and validators for receive invitation request query string."""

    alias = fields.Str(
        required=False,
        metadata={"description": "Alias for connection", "example": "Barry"},
    )
    auto_accept = fields.Boolean(
        required=False,
        metadata={"description": "Auto-accept connection (defaults to configuration)"},
    )
    use_existing_connection = fields.Boolean(
        required=False,
        dump_default=True,
        metadata={"description": "Use an existing connection, if possible"},
    )
    mediation_id = fields.Str(
        required=False,
        validate=UUID4_VALIDATE,
        metadata={
            "description": "Identifier for active mediation record to be used",
            "example": UUID4_EXAMPLE,
        },
    )


class InvitationRecordResponseSchema(OpenAPISchema):
    """Response schema for Invitation Record."""


class InvitationRecordMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking invitation record."""

    invi_msg_id = fields.Str(
        required=True,
        validate=UUID4_VALIDATE,
        metadata={
            "description": "Invitation Message identifier",
            "example": UUID4_EXAMPLE,
        },
    )


class OobInvitationRecordMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking invitation record."""

    oob_id = fields.Str(
        required=True,
        validate=UUID4_VALIDATE,
        metadata={
            "description": "OOB Invitation identifier",
            "example": UUID4_EXAMPLE,
        },
    )


@docs(tags=["out-of-band"], summary="Fetch an existing Out-of-Band invitation.")
@querystring_schema(OobIdQueryStringSchema())
@response_schema(InvitationRecordResponseSchema(), description="")
@tenant_authentication
async def invitation_fetch(request: web.BaseRequest):
    """Request handler for fetching an invitation.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile
    oob_mgr = OutOfBandManager(profile)
    try:
        record = await oob_mgr.fetch_oob_invitation_record_by_id(
            request.query.get("oob_id")
        )
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except StorageError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(record.serialize())


@docs(
    tags=["out-of-band"],
    summary="Create a new connection invitation",
)
@querystring_schema(InvitationCreateQueryStringSchema())
@request_schema(InvitationCreateRequestSchema())
@response_schema(InvitationRecordSchema(), description="")
@tenant_authentication
async def invitation_create(request: web.BaseRequest):
    """Request handler for creating a new connection invitation.

    Args:
        request: aiohttp request object

    Returns:
        The out of band invitation details

    """
    context: AdminRequestContext = request["context"]

    body = await request.json() if request.body_exists else {}
    attachments = body.get("attachments")
    handshake_protocols = body.get("handshake_protocols", [])
    service_accept = body.get("accept")
    use_public_did = body.get("use_public_did", False)
    use_did = body.get("use_did")
    use_did_method = body.get("use_did_method")
    metadata = body.get("metadata")
    my_label = body.get("my_label")
    alias = body.get("alias")
    mediation_id = body.get("mediation_id")
    protocol_version = body.get("protocol_version")
    goal_code = body.get("goal_code")
    goal = body.get("goal")

    multi_use = json.loads(request.query.get("multi_use", "false"))
    auto_accept = json.loads(request.query.get("auto_accept", "null"))
    create_unique_did = json.loads(request.query.get("create_unique_did", "false"))

    profile = context.profile

    oob_mgr = OutOfBandManager(profile)
    try:
        invi_rec = await oob_mgr.create_invitation(
            my_label=my_label,
            auto_accept=auto_accept,
            public=use_public_did,
            use_did=use_did,
            use_did_method=use_did_method,
            hs_protos=[h for h in [HSProto.get(hsp) for hsp in handshake_protocols] if h],
            multi_use=multi_use,
            create_unique_did=create_unique_did,
            attachments=attachments,
            metadata=metadata,
            alias=alias,
            mediation_id=mediation_id,
            service_accept=service_accept,
            protocol_version=protocol_version,
            goal_code=goal_code,
            goal=goal,
        )
    except (StorageNotFoundError, ValidationError, OutOfBandManagerError) as e:
        raise web.HTTPBadRequest(reason=e.roll_up)

    return web.json_response(invi_rec.serialize())


@docs(
    tags=["out-of-band"],
    summary="Receive a new connection invitation",
)
@querystring_schema(InvitationReceiveQueryStringSchema())
@request_schema(InvitationMessageSchema())
@response_schema(OobRecordSchema(), 200, description="")
@tenant_authentication
async def invitation_receive(request: web.BaseRequest):
    """Request handler for receiving a new connection invitation.

    Args:
        request: aiohttp request object

    Returns:
        The out of band invitation details

    """

    context: AdminRequestContext = request["context"]
    if context.settings.get("admin.no_receive_invites"):
        raise web.HTTPForbidden(
            reason="Configuration does not allow receipt of invitations"
        )

    profile = context.profile
    oob_mgr = OutOfBandManager(profile)

    body = await request.json()
    auto_accept = json.loads(request.query.get("auto_accept", "null"))
    alias = request.query.get("alias")
    # By default, try to use an existing connection
    use_existing_conn = json.loads(request.query.get("use_existing_connection", "true"))
    mediation_id = request.query.get("mediation_id")
    try:
        invitation = InvitationMessage.deserialize(body)
        result = await oob_mgr.receive_invitation(
            invitation,
            auto_accept=auto_accept,
            alias=alias,
            use_existing_connection=use_existing_conn,
            mediation_id=mediation_id,
        )
    except (DIDXManagerError, StorageError, BaseModelError) as err:
        LOGGER.exception("Error during receive invitation")
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(result.serialize())


@docs(tags=["out-of-band"], summary="Delete records associated with invitation")
@match_info_schema(InvitationRecordMatchInfoSchema())
@response_schema(InvitationRecordResponseSchema(), description="")
@tenant_authentication
async def invitation_remove(request: web.BaseRequest):
    """Request handler for removing a invitation related conn and oob records.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    invi_msg_id = request.match_info["invi_msg_id"]
    profile = context.profile
    oob_mgr = OutOfBandManager(profile)
    try:
        await oob_mgr.delete_conn_and_oob_record_invitation(invi_msg_id)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except StorageError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({})


async def register(app: web.Application):
    """Register routes."""
    app.add_routes(
        [
            web.post("/out-of-band/create-invitation", invitation_create),
            web.post("/out-of-band/receive-invitation", invitation_receive),
            web.get(
                "/out-of-band/invitations",
                invitation_fetch,
                allow_head=False,
            ),
            web.delete("/out-of-band/invitations/{invi_msg_id}", invitation_remove),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "out-of-band",
            "description": "Out-of-band connections",
            "externalDocs": {
                "description": "Design",
                "url": SPEC_URI,
            },
        }
    )



"""Base configurations and constants for logging."""

DEFAULT_LOGGING_CONFIG_PATH_INI = "acapy_agent.config.logging:default_logging_config.ini"
DEFAULT_MULTITENANT_LOGGING_CONFIG_PATH_INI = (
    "acapy_agent.config.logging:default_multitenant_logging_config.ini"
)
LOG_FORMAT_FILE_ALIAS_PATTERN = (
    "%(asctime)s %(wallet_id)s %(levelname)s %(pathname)s:%(lineno)d %(message)s"
)



"""Plugin hook for providing an external signature suite implementation.

This enables greater control over where JSON-LD credentials are signed without
requiring knowledge of the complexities of the JSON-LD/VC-LDP subsystem.
"""

from abc import ABC, abstractmethod
from typing import Optional

from ...core.error import BaseError
from ...core.profile import Profile
from ...wallet.did_info import DIDInfo
from ..ld_proofs.suites.linked_data_proof import LinkedDataProof


class ExternalSuiteError(BaseError):
    """Raised when an error occurs in an external signature suite provider."""


class ExternalSuiteNotFoundError(ExternalSuiteError):
    """Raised when an external signature suite provider is not found.

    This should be raised to prevent falling back to built in suites, if desired.
    """


class ExternalSuiteProvider(ABC):
    """Plugin hook for providing an external signature suite implementation."""

    @abstractmethod
    async def get_suite(
        self,
        profile: Profile,
        proof_type: str,
        proof: dict,
        verification_method: str,
        did_info: DIDInfo,
    ) -> Optional[LinkedDataProof]:
        """Get a signature suite for the given proof type and verification method.

        Implementing classes should raise ExternalSuiteNotFoundError if preventing
        fallback to built-in suites is desired. Otherwise, return None to indicate
        that the implementing class does not support the given proof type.
        """



from unittest import IsolatedAsyncioTestCase

from ...askar.profile import AskarProfileSession
from ...utils.stats import Collector
from ...utils.testing import create_test_profile
from .. import request_context as test_module


class TestAdminRequestContext(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        self.ctx = test_module.AdminRequestContext(self.profile)
        assert self.ctx.__class__.__name__ in str(self.ctx)

        self.ctx_with_added_attrs = test_module.AdminRequestContext(
            profile=self.profile,
            metadata={"test_attrib_key": "test_attrib_value"},
        )
        assert self.ctx_with_added_attrs.__class__.__name__ in str(
            self.ctx_with_added_attrs
        )

    def test_session_transaction(self):
        sesn = self.ctx.session()
        assert isinstance(sesn, AskarProfileSession)
        txn = self.ctx.transaction()
        assert isinstance(txn, AskarProfileSession)

        sesn = self.ctx_with_added_attrs.session()
        assert isinstance(sesn, AskarProfileSession)
        txn = self.ctx_with_added_attrs.transaction()
        assert isinstance(txn, AskarProfileSession)

    async def test_session_inject_x(self):
        test_ctx = test_module.AdminRequestContext(self.profile)
        async with test_ctx.session() as session:
            with self.assertRaises(test_module.InjectionError):
                session.inject(Collector)



"""Connection complete handler under RFC 23 (DID exchange)."""

from .....messaging.base_handler import BaseHandler, BaseResponder, RequestContext
from ..manager import DIDXManager, DIDXManagerError
from ..messages.complete import DIDXComplete


class DIDXCompleteHandler(BaseHandler):
    """Handler class for connection complete message under RFC 23 (DID exchange)."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle connection complete under RFC 23 (DID exchange).

        Args:
            context: Request context
            responder: Responder callback
        """
        self._logger.debug("DIDXCompleteHandler called with context %s", context)
        assert isinstance(context.message, DIDXComplete)

        profile = context.profile
        mgr = DIDXManager(profile)
        try:
            await mgr.accept_complete(context.message, context.message_receipt)
        except DIDXManagerError:
            # no corresponding request: no targets to send problem report; log and quit
            self._logger.exception("Error receiving connection complete")



DID_SOV_QqEfJxe752NCmWqR5TssZ5 = {
    "@context": "https://www.w3.org/ns/did/v1",
    "id": "did:sov:QqEfJxe752NCmWqR5TssZ5",
    "verificationMethod": [
        {
            "id": "did:sov:QqEfJxe752NCmWqR5TssZ5#key-1",
            "type": "Ed25519VerificationKey2018",
            "controller": "did:sov:QqEfJxe752NCmWqR5TssZ5",
            "publicKeyBase58": "DzNC1pbarUzgGXmxRsccNJDBjWgCaiy6uSXgPPJZGWCL",
        }
    ],
    "authentication": ["did:sov:QqEfJxe752NCmWqR5TssZ5#key-1"],
    "assertionMethod": ["did:sov:QqEfJxe752NCmWqR5TssZ5#key-1"],
    "service": [
        {
            "id": "did:sov:QqEfJxe752NCmWqR5TssZ5#did-communication",
            "type": "did-communication",
            "serviceEndpoint": "http://localhost:3002",
            "recipientKeys": ["did:sov:QqEfJxe752NCmWqR5TssZ5#key-1"],
            "routingKeys": [],
            "priority": 0,
        }
    ],
}



"""Peer DID Resolver.

Resolution is performed using the peer-did-python library https://github.com/sicpa-dlab/peer-did-python.
"""

from typing import Optional, Pattern, Sequence, Text

from did_peer_2 import PATTERN, resolve

from ...config.injection_context import InjectionContext
from ...core.profile import Profile
from ..base import BaseDIDResolver, DIDNotFound, ResolverType
from .peer3 import PeerDID3Resolver


class PeerDID2Resolver(BaseDIDResolver):
    """Peer DID Resolver."""

    def __init__(self):
        """Initialize Key Resolver."""
        super().__init__(ResolverType.NATIVE)

    async def setup(self, context: InjectionContext):
        """Perform required setup for Key DID resolution."""

    @property
    def supported_did_regex(self) -> Pattern:
        """Return supported_did_regex of Key DID Resolver."""
        return PATTERN

    async def _resolve(
        self,
        profile: Profile,
        did: str,
        service_accept: Optional[Sequence[Text]] = None,
    ) -> dict:
        """Resolve a Key DID."""
        if not PATTERN.match(did):
            raise DIDNotFound(f"did is not a peer did: {did}")

        doc = resolve(did)
        await PeerDID3Resolver().create_and_store(profile, did)

        return doc



"""Verifiable Credential and Presentation verification methods."""

from typing import Optional

from anoncreds import AnoncredsError

from acapy_agent.anoncreds.holder import AnonCredsHolderError

from ...anoncreds.verifier import AnonCredsVerifier
from ...core.profile import Profile
from ..ld_proofs import ProofPurpose
from ..vc_ld.validation_result import PresentationVerificationResult
from .prove import _load_w3c_credentials, prepare_data_for_presentation


async def verify_signed_anoncredspresentation(
    *,
    profile: Profile,
    presentation: dict,
    purpose: Optional[ProofPurpose] = None,
    challenge: Optional[str] = None,
    domain: Optional[str] = None,
    pres_req: Optional[dict] = None,
) -> PresentationVerificationResult:
    """Verify presentation structure, credentials, proof purpose and signature.

    Args:
        profile (Profile): The profile to use for verification
        presentation (dict): The presentation to verify
        purpose (ProofPurpose, optional): Proof purpose to use.
        challenge (str, optional): The challenge to use for authentication.
        domain (str, optional): Domain to use for the authentication proof purpose.
        pres_req (dict, optional): The presentation request to verify against.

    Returns:
        PresentationVerificationResult: The result of the verification. Verified property
            indicates whether the verification was successful

    """

    # TODO: I think we should add some sort of options to authenticate the subject id
    # to the presentation verification method controller
    anoncreds_verifier = AnonCredsVerifier(profile)

    credentials = presentation["verifiableCredential"]
    pres_definition = pres_req["presentation_definition"]
    w3c_creds = await _load_w3c_credentials(credentials)

    (anoncreds_pres_req, cred_metadata) = await prepare_data_for_presentation(
        presentation=presentation,
        w3c_creds=w3c_creds,
        pres_definition=pres_definition,
        profile=profile,
        challenge=challenge,
    )

    try:
        return await anoncreds_verifier.verify_presentation_w3c(
            anoncreds_pres_req, presentation, cred_metadata
        )
    except AnoncredsError as err:
        raise AnonCredsHolderError("Error loading master secret") from err



"""Test Problem Report Handler."""

import pytest
import pytest_asyncio

from ......connections.models.conn_record import ConnRecord
from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......tests import mock
from ......transport.inbound.receipt import MessageReceipt
from ......utils.testing import create_test_profile
from ...handlers import problem_report_handler as test_module
from ...manager import OutOfBandManagerError
from ...messages.problem_report import OOBProblemReport, ProblemReportReason


@pytest_asyncio.fixture
async def request_context():
    ctx = RequestContext.test_context(await create_test_profile())
    ctx.message_receipt = MessageReceipt()
    yield ctx


@pytest_asyncio.fixture
async def connection_record(request_context, session):
    record = ConnRecord()
    request_context.connection_record = record
    await record.save(session)
    yield record


@pytest_asyncio.fixture
async def session(request_context):
    yield await request_context.session()


class TestOOBProblemReportHandler:
    @pytest.mark.asyncio
    @mock.patch.object(test_module, "OutOfBandManager")
    async def test_called(self, mock_oob_mgr, request_context, connection_record):
        mock_oob_mgr.return_value.receive_problem_report = mock.CoroutineMock()
        request_context.message = OOBProblemReport(
            description={
                "en": "No such connection",
                "code": ProblemReportReason.NO_EXISTING_CONNECTION.value,
            }
        )
        handler = test_module.OOBProblemReportMessageHandler()
        responder = MockResponder()
        await handler.handle(context=request_context, responder=responder)
        mock_oob_mgr.return_value.receive_problem_report.assert_called_once_with(
            problem_report=request_context.message,
            receipt=request_context.message_receipt,
            conn_record=connection_record,
        )

    @pytest.mark.asyncio
    @mock.patch.object(test_module, "OutOfBandManager")
    async def test_exception(self, mock_oob_mgr, request_context, connection_record):
        mock_oob_mgr.return_value.receive_problem_report = mock.CoroutineMock()
        mock_oob_mgr.return_value.receive_problem_report.side_effect = (
            OutOfBandManagerError("error")
        )
        request_context.message = OOBProblemReport(
            description={
                "en": "Connection not active",
                "code": ProblemReportReason.EXISTING_CONNECTION_NOT_ACTIVE.value,
            }
        )
        handler = test_module.OOBProblemReportMessageHandler()
        with mock.patch.object(
            handler._logger, "exception", mock.MagicMock()
        ) as mock_exc_logger:
            responder = MockResponder()
            await handler.handle(context=request_context, responder=responder)

        mock_exc_logger.assert_called_once()



"""Record used to handle routing of messages to another agent."""

from typing import Optional, Sequence

from marshmallow import EXCLUDE, fields

from ...messaging.models.base import BaseModel, BaseModelSchema
from ...messaging.valid import (
    GENERIC_DID_EXAMPLE,
    GENERIC_DID_VALIDATE,
    RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
    RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
)


class ConnectionTarget(BaseModel):
    """Record used to handle routing of messages to another agent."""

    class Meta:
        """ConnectionTarget metadata."""

        schema_class = "ConnectionTargetSchema"

    def __init__(
        self,
        *,
        did: Optional[str] = None,
        endpoint: Optional[str] = None,
        label: Optional[str] = None,
        recipient_keys: Optional[Sequence[str]] = None,
        routing_keys: Optional[Sequence[str]] = None,
        sender_key: Optional[str] = None,
    ):
        """Initialize a ConnectionTarget instance.

        Args:
            did: A did for the connection
            endpoint: An endpoint for the connection
            label: A label for the connection
            recipient_keys: A list of recipient keys
            routing_keys: A list of routing keys
            sender_key: A sender key
        """
        self.did = did
        self.endpoint = endpoint
        self.label = label
        self.recipient_keys = list(recipient_keys) if recipient_keys else []
        self.routing_keys = list(routing_keys) if routing_keys else []
        self.sender_key = sender_key


class ConnectionTargetSchema(BaseModelSchema):
    """ConnectionTarget schema."""

    class Meta:
        """ConnectionTargetSchema metadata."""

        model_class = ConnectionTarget
        unknown = EXCLUDE

    did = fields.Str(
        required=False,
        validate=GENERIC_DID_VALIDATE,
        metadata={"description": "", "example": GENERIC_DID_EXAMPLE},
    )
    endpoint = fields.Str(
        required=False,
        metadata={
            "description": "Connection endpoint",
            "example": "http://192.168.56.102:8020",
        },
    )
    label = fields.Str(
        required=False, metadata={"description": "Connection label", "example": "Bob"}
    )
    recipient_keys = fields.List(
        fields.Str(
            validate=RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
            metadata={
                "description": "Recipient public key",
                "example": RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
            },
        ),
        required=False,
        metadata={"description": "List of recipient keys"},
    )
    routing_keys = fields.List(
        fields.Str(
            validate=RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
            metadata={
                "description": "Routing key",
                "example": RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
            },
        ),
        data_key="routingKeys",
        required=False,
        metadata={"description": "List of routing keys"},
    )
    sender_key = fields.Str(
        required=False,
        validate=RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
        metadata={
            "description": "Sender public key",
            "example": RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
        },
    )



"""Automated setup process for AnonCreds credential definitions with revocation."""

import logging
from abc import ABC, abstractmethod

from acapy_agent.protocols.endorse_transaction.v1_0.util import is_author_role

from ..anoncreds.revocation import AnonCredsRevocation, AnonCredsRevocationError
from ..core.event_bus import EventBus
from ..core.profile import Profile
from ..revocation.util import notify_revocation_published_event
from .events import (
    CRED_DEF_FINISHED_PATTERN,
    REV_LIST_FINISHED_PATTERN,
    REV_REG_DEF_FINISHED_PATTERN,
    CredDefFinishedEvent,
    RevListFinishedEvent,
    RevRegDefFinishedEvent,
)

LOGGER = logging.getLogger(__name__)


class AnonCredsRevocationSetupManager(ABC):
    """Base class for automated setup of revocation."""

    @abstractmethod
    def register_events(self, event_bus: EventBus):
        """Event registration."""


class DefaultRevocationSetup(AnonCredsRevocationSetupManager):
    """Manager for automated setup of revocation support.

    This manager models a state machine for the revocation setup process where
    the transitions are triggered by the `finished` event of the previous
    artifact. The state machine is as follows:

    [*] --> Cred Def
    Cred Def --> Rev Reg Def
    Rev Reg Def --> Rev List
    Rev List --> [*]

    This implementation of an AnonCredsRevocationSetupManager will create two
    revocation registries for each credential definition supporting revocation;
    one that is active and one that is pending. When the active registry fills,
    the pending registry will be activated and a new pending registry will be
    created. This will continue indefinitely.

    This hot-swap approach to revocation registry management allows for
    issuance operations to be performed without a delay for registry
    creation.
    """

    REGISTRY_TYPE = "CL_ACCUM"
    INITIAL_REGISTRY_COUNT = 2

    def __init__(self):
        """Init manager."""

    def register_events(self, event_bus: EventBus) -> None:
        """Register event listeners."""
        event_bus.subscribe(CRED_DEF_FINISHED_PATTERN, self.on_cred_def)
        event_bus.subscribe(REV_REG_DEF_FINISHED_PATTERN, self.on_rev_reg_def)
        event_bus.subscribe(REV_LIST_FINISHED_PATTERN, self.on_rev_list)

    async def on_cred_def(self, profile: Profile, event: CredDefFinishedEvent) -> None:
        """Handle cred def finished."""
        payload = event.payload

        if payload.support_revocation:
            revoc = AnonCredsRevocation(profile)
            for registry_count in range(self.INITIAL_REGISTRY_COUNT):
                await revoc.create_and_register_revocation_registry_definition(
                    issuer_id=payload.issuer_id,
                    cred_def_id=payload.cred_def_id,
                    registry_type=self.REGISTRY_TYPE,
                    max_cred_num=payload.max_cred_num,
                    tag=str(registry_count),
                    options=payload.options,
                )

    async def on_rev_reg_def(
        self, profile: Profile, event: RevRegDefFinishedEvent
    ) -> None:
        """Handle rev reg def finished."""
        payload = event.payload

        auto_create_revocation = True
        if is_author_role(profile):
            auto_create_revocation = profile.settings.get(
                "endorser.auto_create_rev_reg", False
            )

        if auto_create_revocation:
            revoc = AnonCredsRevocation(profile)
            failed_to_upload_tails = False
            try:
                await revoc.upload_tails_file(payload.rev_reg_def)
            except AnonCredsRevocationError as err:
                LOGGER.warning(f"Failed to upload tails file: {err}")
                failed_to_upload_tails = True

            if failed_to_upload_tails:
                payload.options["failed_to_upload"] = True

            await revoc.create_and_register_revocation_list(
                payload.rev_reg_def_id, payload.options
            )

            if payload.rev_reg_def.tag == str(0):
                # Mark the first registry as active
                await revoc.set_active_registry(payload.rev_reg_def_id)

    async def on_rev_list(self, profile: Profile, event: RevListFinishedEvent) -> None:
        """Handle rev list finished."""
        await notify_revocation_published_event(
            profile, event.payload.rev_reg_id, event.payload.revoked
        )



"""Aries-Askar implementation of BaseWallet interface."""

import asyncio
import json
import logging
from typing import List, Optional, Sequence, Tuple, Union, cast

from aries_askar import AskarError, AskarErrorCode, Entry, Key, KeyAlg, SeedMethod

from ..askar.didcomm.v1 import pack_message, unpack_message
from ..askar.profile import AskarProfileSession
from ..ledger.base import BaseLedger
from ..ledger.endpoint_type import EndpointType
from ..ledger.error import LedgerConfigError
from ..storage.askar import AskarStorage
from ..storage.base import StorageDuplicateError, StorageNotFoundError, StorageRecord
from .base import BaseWallet, DIDInfo, KeyInfo
from .crypto import sign_message, validate_seed, verify_signed_message
from .did_info import INVITATION_REUSE_KEY
from .did_method import INDY, SOV, DIDMethod, DIDMethods
from .did_parameters_validation import DIDParametersValidation
from .error import WalletDuplicateError, WalletError, WalletNotFoundError
from .key_type import BLS12381G2, ED25519, P256, X25519, KeyType, KeyTypes
from .keys.manager import verkey_to_multikey
from .util import b58_to_bytes, bytes_to_b58

CATEGORY_DID = "did"
CATEGORY_CONFIG = "config"
RECORD_NAME_PUBLIC_DID = "default_public_did"

LOGGER = logging.getLogger(__name__)


class AskarWallet(BaseWallet):
    """Aries-Askar wallet implementation."""

    def __init__(self, session: AskarProfileSession):
        """Initialize a new `AskarWallet` instance.

        Args:
            session: The Askar profile session instance to use
        """
        self._session = session

    @property
    def session(self) -> AskarProfileSession:
        """Accessor for Askar profile session instance."""
        return self._session

    async def create_signing_key(
        self,
        key_type: KeyType,
        seed: Optional[str] = None,
        metadata: Optional[dict] = None,
    ) -> KeyInfo:
        """Create a new public/private signing keypair.

        Args:
            key_type: Key type to create
            seed: Seed for key
            metadata: Optional metadata to store with the keypair

        Returns:
            A `KeyInfo` representing the new record

        Raises:
            WalletDuplicateError: If the resulting verkey already exists in the wallet
            WalletError: If there is another backend error

        """
        return await self.create_key(key_type, seed, metadata)

    async def create_key(
        self,
        key_type: KeyType,
        seed: Optional[str] = None,
        metadata: Optional[dict] = None,
        kid: Optional[str] = None,
    ) -> KeyInfo:
        """Create a new public/private keypair.

        Args:
            key_type: Key type to create
            seed: Seed for key
            metadata: Optional metadata to store with the keypair
            kid: Optional key identifier

        Returns:
            A `KeyInfo` representing the new record

        Raises:
            WalletDuplicateError: If the resulting verkey already exists in the wallet
            WalletError: If there is another backend error
        """
        if metadata is None:
            metadata = {}

        try:
            keypair = _create_keypair(key_type, seed)
            verkey = bytes_to_b58(keypair.get_public_bytes())
            tags = {
                "multikey": verkey_to_multikey(verkey, key_type.key_type),
                "kid": [kid] if kid else [],
            }

            await self._session.handle.insert_key(
                verkey,
                keypair,
                metadata=json.dumps(metadata),
                tags=tags,
            )
        except AskarError as err:
            if err.code == AskarErrorCode.DUPLICATE:
                raise WalletDuplicateError(
                    "Verification key already present in wallet"
                ) from None
            raise WalletError("Error creating signing key") from err
        return KeyInfo(verkey=verkey, metadata=metadata, key_type=key_type, kid=kid)

    async def assign_kid_to_key(self, verkey: str, kid: str) -> KeyInfo:
        """Assign a KID to a key.

        This is separate from the create_key method because some DIDs are only
        known after keys are created.

        Args:
            verkey: The verification key of the keypair
            kid: The kid to assign to the keypair

        Returns:
            A `KeyInfo` representing the keypair

        """
        key_entry = await self._session.handle.fetch_key(name=verkey, for_update=True)
        if not key_entry:
            raise WalletNotFoundError(f"No key entry found for verkey {verkey}")

        key = cast(Key, key_entry.key)
        metadata = cast(dict, key_entry.metadata)
        key_types = self.session.inject(KeyTypes)
        key_type = key_types.from_key_type(key.algorithm.value)
        if not key_type:
            raise WalletError(f"Unknown key type {key.algorithm.value}")

        tags = key_entry.tags or {"kid": []}
        key_ids = tags.get("kid", [])
        key_ids = key_ids if isinstance(key_ids, list) else [key_ids]
        key_ids.append(kid)
        tags["kid"] = key_ids

        await self._session.handle.update_key(name=verkey, tags=tags)
        return KeyInfo(verkey=verkey, metadata=metadata, key_type=key_type, kid=kid)

    async def unassign_kid_from_key(self, verkey: str, kid: str) -> KeyInfo:
        """Remove a kid association.

        Args:
            kid: The key identifier
            verkey: The verification key of the keypair

        Returns:
            The key identified by kid

        """
        key_entries = await self._session.handle.fetch_all_keys(
            tag_filter={"kid": kid}, limit=2
        )
        if len(key_entries) > 1:
            raise WalletDuplicateError(f"More than one key found by kid {kid}")

        key_entry = key_entries[0]
        key = cast(Key, key_entry.key)
        fetched_verkey = bytes_to_b58(key.get_public_bytes())

        metadata = cast(dict, key_entry.metadata)
        key_types = self.session.inject(KeyTypes)
        key_type = key_types.from_key_type(key.algorithm.value)
        if not key_type:
            raise WalletError(f"Unknown key type {key.algorithm.value}")

        if fetched_verkey != verkey:
            raise WalletError(f"Multikey mismatch: {fetched_verkey} != {verkey}")

        key_tags = key_entry.tags or {"kid": []}
        key_kids = key_tags.get("kid", [])
        key_kids = key_kids if isinstance(key_kids, list) else [key_kids]

        try:
            key_kids.remove(kid)
        except ValueError:
            pass

        key_tags["kid"] = key_kids

        await self._session.handle.update_key(name=verkey, tags=key_tags)
        return KeyInfo(verkey=verkey, metadata=metadata, key_type=key_type, kid=kid)

    async def get_key_by_kid(self, kid: str) -> KeyInfo:
        """Fetch a key by looking up its kid.

        Args:
            kid: the key identifier

        Returns:
            The key identified by kid

        """
        key_entries = await self._session.handle.fetch_all_keys(
            tag_filter={"kid": kid}, limit=2
        )
        if len(key_entries) > 1:
            raise WalletDuplicateError(f"More than one key found by kid {kid}")
        elif len(key_entries) < 1:
            raise WalletNotFoundError(f"No key found for kid {kid}")

        entry = key_entries[0]
        key = cast(Key, entry.key)
        verkey = bytes_to_b58(key.get_public_bytes())
        metadata = cast(dict, entry.metadata)
        key_types = self.session.inject(KeyTypes)
        key_type = key_types.from_key_type(key.algorithm.value)
        if not key_type:
            raise WalletError(f"Unknown key type {key.algorithm.value}")

        return KeyInfo(verkey=verkey, metadata=metadata, key_type=key_type, kid=kid)

    async def get_signing_key(self, verkey: str) -> KeyInfo:
        """Fetch info for a signing keypair.

        Args:
            verkey: The verification key of the keypair

        Returns:
            A `KeyInfo` representing the keypair

        Raises:
            WalletNotFoundError: If no keypair is associated with the verification key
            WalletError: If there is another backend error

        """

        if not verkey:
            raise WalletNotFoundError("No key identifier provided")
        key_entry = await self._session.handle.fetch_key(verkey)
        if not key_entry:
            raise WalletNotFoundError("Unknown key: {}".format(verkey))
        metadata = json.loads(key_entry.metadata or "{}")

        kid = key_entry.tags.get("kid", []) if key_entry.tags else []

        key = cast(Key, key_entry.key)
        key_types = self.session.inject(KeyTypes)
        key_type = key_types.from_key_type(key.algorithm.value)
        if not key_type:
            raise WalletError(f"Unknown key type {key.algorithm.value}")
        return KeyInfo(verkey=verkey, metadata=metadata, key_type=key_type, kid=kid)

    async def replace_signing_key_metadata(self, verkey: str, metadata: dict):
        """Replace the metadata associated with a signing keypair.

        Args:
            verkey: The verification key of the keypair
            metadata: The new metadata to store

        Raises:
            WalletNotFoundError: if no keypair is associated with the verification key

        """

        # FIXME caller should always create a transaction first

        if not verkey:
            raise WalletNotFoundError("No key identifier provided")

        key = await self._session.handle.fetch_key(verkey, for_update=True)
        if not key:
            raise WalletNotFoundError("Keypair not found")
        await self._session.handle.update_key(
            verkey, metadata=json.dumps(metadata or {}), tags=key.tags
        )

    async def create_local_did(
        self,
        method: DIDMethod,
        key_type: KeyType,
        seed: Optional[str] = None,
        did: Optional[str] = None,
        metadata: Optional[dict] = None,
    ) -> DIDInfo:
        """Create and store a new local DID.

        Args:
            method: The method to use for the DID
            key_type: The key type to use for the DID
            seed: Optional seed to use for DID
            did: The DID to use
            metadata: Metadata to store with DID

        Returns:
            A `DIDInfo` instance representing the created DID

        Raises:
            WalletDuplicateError: If the DID already exists in the wallet
            WalletError: If there is another backend error

        """
        LOGGER.debug(
            "Creating local %s %s DID %s%s",
            method.method_name,
            key_type.key_type,
            did or "",
            " from seed" if seed else "",
        )
        did_validation = DIDParametersValidation(self._session.context.inject(DIDMethods))
        did_validation.validate_key_type(method, key_type)

        if not metadata:
            metadata = {}

        try:
            keypair = _create_keypair(key_type, seed)
            verkey_bytes = keypair.get_public_bytes()
            verkey = bytes_to_b58(verkey_bytes)

            did = did_validation.validate_or_derive_did(
                method, key_type, verkey_bytes, did
            )

            try:
                await self._session.handle.insert_key(
                    verkey, keypair, metadata=json.dumps(metadata)
                )
            except AskarError as err:
                if err.code == AskarErrorCode.DUPLICATE:
                    # update metadata?
                    pass
                else:
                    raise WalletError("Error inserting key") from err

            item = await self._session.handle.fetch(CATEGORY_DID, did, for_update=True)
            if item:
                did_info = item.value_json
                if did_info.get("verkey") != verkey:
                    raise WalletDuplicateError("DID already present in wallet")
                if did_info.get("metadata") != metadata:
                    did_info["metadata"] = metadata
                    await self._session.handle.replace(
                        CATEGORY_DID, did, value_json=did_info, tags=item.tags
                    )
            else:
                value_json = {
                    "did": did,
                    "method": method.method_name,
                    "verkey": verkey,
                    "verkey_type": key_type.key_type,
                    "metadata": metadata,
                }
                tags = {
                    "method": method.method_name,
                    "verkey": verkey,
                    "verkey_type": key_type.key_type,
                }
                if INVITATION_REUSE_KEY in metadata:
                    tags[INVITATION_REUSE_KEY] = "true"
                await self._session.handle.insert(
                    CATEGORY_DID,
                    did,
                    value_json=value_json,
                    tags=tags,
                )

        except AskarError as err:
            raise WalletError("Error when creating local DID") from err

        return DIDInfo(
            did=did, verkey=verkey, metadata=metadata, method=method, key_type=key_type
        )

    async def store_did(self, did_info: DIDInfo) -> DIDInfo:
        """Store a DID in the wallet.

        This enables components external to the wallet to define how a DID
        is created and then store it in the wallet for later use.

        Args:
            did_info: The DID to store

        Returns:
            The stored `DIDInfo`
        """
        LOGGER.debug("Storing DID %s", did_info.did)
        try:
            item = await self._session.handle.fetch(
                CATEGORY_DID, did_info.did, for_update=True
            )
            if item:
                raise WalletDuplicateError("DID already present in wallet")
            else:
                value_json = {
                    "did": did_info.did,
                    "method": did_info.method.method_name,
                    "verkey": did_info.verkey,
                    "verkey_type": did_info.key_type.key_type,
                    "metadata": did_info.metadata,
                }
                tags = {
                    "method": did_info.method.method_name,
                    "verkey": did_info.verkey,
                    "verkey_type": did_info.key_type.key_type,
                }
                if INVITATION_REUSE_KEY in did_info.metadata:
                    tags[INVITATION_REUSE_KEY] = "true"
                await self._session.handle.insert(
                    CATEGORY_DID,
                    did_info.did,
                    value_json=value_json,
                    tags=tags,
                )
        except AskarError as err:
            raise WalletError("Error when storing DID") from err

        return did_info

    async def get_local_dids(self) -> Sequence[DIDInfo]:
        """Get list of defined local DIDs.

        Returns:
            A list of locally stored DIDs as `DIDInfo` instances

        """
        LOGGER.debug("Getting local DIDs")
        ret = []
        for item in await self._session.handle.fetch_all(CATEGORY_DID):
            ret.append(self._load_did_entry(item))
        return ret

    async def get_local_did(self, did: str) -> DIDInfo:
        """Find info for a local DID.

        Args:
            did: The DID for which to get info

        Returns:
            A `DIDInfo` instance representing the found DID

        Raises:
            WalletNotFoundError: If the DID is not found
            WalletError: If there is another backend error

        """
        LOGGER.debug("Getting local DID for DID %s", did)
        if not did:
            raise WalletNotFoundError("No identifier provided")
        try:
            did_entry = await self._session.handle.fetch(CATEGORY_DID, did)
        except AskarError as err:
            raise WalletError("Error when fetching local DID") from err
        if not did_entry:
            raise WalletNotFoundError("Unknown DID: {}".format(did))
        return self._load_did_entry(did_entry)

    async def get_local_did_for_verkey(self, verkey: str) -> DIDInfo:
        """Resolve a local DID from a verkey.

        Args:
            verkey: The verkey for which to get the local DID

        Returns:
            A `DIDInfo` instance representing the found DID

        Raises:
            WalletNotFoundError: If the verkey is not found

        """
        LOGGER.debug("Getting local DID for verkey %s", verkey)
        try:
            dids = await self._session.handle.fetch_all(CATEGORY_DID, {"verkey": verkey})
        except AskarError as err:
            raise WalletError("Error when fetching local DID for verkey") from err
        if dids:
            ret_did = dids[0]
            ret_did_info = ret_did.value_json
            if len(dids) > 1 and ret_did_info["did"].startswith("did:peer:4"):
                # if it is a peer:did:4 make sure we are using the short version
                other_did = dids[1]  # assume only 2
                other_did_info = other_did.value_json
                if len(other_did_info["did"]) < len(ret_did_info["did"]):
                    ret_did = other_did
                    ret_did_info = other_did.value_json
            return self._load_did_entry(ret_did)
        raise WalletNotFoundError("No DID defined for verkey: {}".format(verkey))

    async def replace_local_did_metadata(self, did: str, metadata: dict):
        """Replace metadata for a local DID.

        Args:
            did: The DID for which to replace metadata
            metadata: The new metadata

        """
        LOGGER.debug("Replacing metadata for DID %s with %s", did, metadata)

        try:
            item = await self._session.handle.fetch(CATEGORY_DID, did, for_update=True)
            if not item:
                LOGGER.warning("DID %s not found when replacing metadata", did)
                raise WalletNotFoundError("Unknown DID: {}".format(did)) from None
            entry_val = item.value_json
            if entry_val["metadata"] != metadata:
                entry_val["metadata"] = metadata
                await self._session.handle.replace(
                    CATEGORY_DID, did, value_json=entry_val, tags=item.tags
                )
        except AskarError as err:
            LOGGER.error("Error updating DID metadata: %s", err)
            raise WalletError("Error updating DID metadata") from err

    async def get_public_did(self) -> DIDInfo | None:
        """Retrieve the public DID.

        Returns:
            The currently public `DIDInfo`, if any

        """
        LOGGER.debug("Retrieving public DID")
        public_did = None
        public_info = None
        public_item = None
        storage = AskarStorage(self._session)
        try:
            public_item = await storage.get_record(
                CATEGORY_CONFIG, RECORD_NAME_PUBLIC_DID
            )
        except StorageNotFoundError:
            # populate public DID record
            # this should only happen once, for an upgraded wallet
            # the 'public' metadata flag is no longer used
            LOGGER.debug("No %s found, retrieving local DIDs", RECORD_NAME_PUBLIC_DID)
            dids = await self.get_local_dids()
            for info in dids:
                if info.metadata.get("public"):
                    public_did = info.did
                    public_info = info
                    LOGGER.debug("Public DID found: %s", public_did)
                    break
            try:
                # even if public is not set, store a record
                # to avoid repeated queries
                LOGGER.debug("Adding %s record", RECORD_NAME_PUBLIC_DID)
                await storage.add_record(
                    StorageRecord(
                        type=CATEGORY_CONFIG,
                        id=RECORD_NAME_PUBLIC_DID,
                        value=json.dumps({"did": public_did}),
                    )
                )
            except StorageDuplicateError:
                LOGGER.debug(
                    "Another process stored the %s record first", RECORD_NAME_PUBLIC_DID
                )
                public_item = await storage.get_record(
                    CATEGORY_CONFIG, RECORD_NAME_PUBLIC_DID
                )
        if public_item:
            LOGGER.debug("Public DID storage record found")
            public_did = json.loads(public_item.value)["did"]
            if public_did:
                try:
                    public_info = await self.get_local_did(public_did)
                    LOGGER.debug("Public DID found in wallet: %s", public_did)
                except WalletNotFoundError:
                    LOGGER.debug("Public DID not found in wallet: %s", public_did)
            else:
                LOGGER.debug("DID not found in public DID storage record: %s", public_did)

        return public_info

    async def set_public_did(self, did: Union[str, DIDInfo]) -> DIDInfo:
        """Assign the public DID.

        Returns:
            The updated `DIDInfo`

        """

        if isinstance(did, str):
            try:
                item = await self._session.handle.fetch(
                    CATEGORY_DID, did, for_update=True
                )
            except AskarError as err:
                raise WalletError("Error when fetching local DID") from err
            if not item:
                raise WalletNotFoundError("Unknown DID: {}".format(did))
            info = self._load_did_entry(item)
        else:
            info = did
            item = None

        public = await self.get_public_did()
        if not public or public.did != info.did:
            storage = AskarStorage(self._session)
            if not info.metadata.get("posted"):
                LOGGER.debug("Setting posted flag for DID %s", info.did)
                metadata = {**info.metadata, "posted": True}
                if item:
                    entry_val = item.value_json
                    entry_val["metadata"] = metadata
                    await self._session.handle.replace(
                        CATEGORY_DID, did, value_json=entry_val, tags=item.tags
                    )
                else:
                    await self.replace_local_did_metadata(info.did, metadata)
                info = info._replace(
                    metadata=metadata,
                )
            LOGGER.debug("Updating public DID to %s", info.did)
            await storage.update_record(
                StorageRecord(
                    type=CATEGORY_CONFIG,
                    id=RECORD_NAME_PUBLIC_DID,
                    value="{}",
                ),
                value=json.dumps({"did": info.did}),
                tags=None,
            )
            public = info
        else:
            LOGGER.warning("Public DID is already set to %s", public.did)

        return public

    async def set_did_endpoint(
        self,
        did: str,
        endpoint: str,
        ledger: BaseLedger,
        endpoint_type: Optional[EndpointType] = None,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
        routing_keys: Optional[List[str]] = None,
    ):
        """Update the endpoint for a DID in the wallet, send to ledger if posted.

        Args:
            did (str): The DID for which to set the endpoint.
            endpoint (str): The endpoint to set. Use None to clear the endpoint.
            ledger (BaseLedger): The ledger to which to send the endpoint update if the
                DID is public or posted.
            endpoint_type (EndpointType, optional): The type of the endpoint/service.
                Only endpoint_type 'endpoint' affects the local wallet. Defaults to None.
            write_ledger (bool, optional): Whether to write the endpoint update to the
                ledger. Defaults to True.
            endorser_did (str, optional): The DID of the endorser. Defaults to None.
            routing_keys (List[str], optional): The routing keys to be used.
                Defaults to None.

        Raises:
            WalletError: If the DID is not of type 'did:sov'.
            LedgerConfigError: If no ledger is available but the DID is public.

        Returns:
            dict: The attribute definition if write_ledger is False, otherwise None.

        """
        LOGGER.debug("Setting endpoint for DID %s to %s", did, endpoint)
        did_info = await self.get_local_did(did)
        if did_info.method not in (SOV, INDY):
            raise WalletError(
                "Setting DID endpoint is only allowed for did:sov or did:indy DIDs"
            )
        metadata = {**did_info.metadata}
        if not endpoint_type:
            endpoint_type = EndpointType.ENDPOINT
        if endpoint_type == EndpointType.ENDPOINT:
            metadata[endpoint_type.indy] = endpoint

        wallet_public_didinfo = await self.get_public_did()
        if (
            wallet_public_didinfo and wallet_public_didinfo.did == did
        ) or did_info.metadata.get("posted"):
            # if DID on ledger, set endpoint there first
            if not ledger:
                LOGGER.error("No ledger available but DID %s is public", did)
                raise LedgerConfigError(
                    f"No ledger available but DID {did} is public: missing wallet-type?"
                )
            if not ledger.read_only:
                LOGGER.debug("Updating endpoint for DID %s on ledger", did)
                async with ledger:
                    attrib_def = await ledger.update_endpoint_for_did(
                        did,
                        endpoint,
                        endpoint_type,
                        write_ledger=write_ledger,
                        endorser_did=endorser_did,
                        routing_keys=routing_keys,
                    )
                    if not write_ledger:
                        return attrib_def

        await self.replace_local_did_metadata(did, metadata)

    async def rotate_did_keypair_start(
        self, did: str, next_seed: Optional[str] = None
    ) -> str:
        """Begin key rotation for DID that wallet owns: generate new keypair.

        Args:
            did: signing DID
            next_seed: incoming replacement seed (default random)

        Returns:
            The new verification key

        """
        # Check if DID can rotate keys
        did_methods = self._session.inject(DIDMethods)
        did_method: DIDMethod = did_methods.from_did(did)
        if not did_method.supports_rotation:
            raise WalletError(
                f"DID method '{did_method.method_name}' does not support key rotation."
            )

        # create a new key to be rotated to (only did:sov/ED25519 supported for now)
        keypair = _create_keypair(ED25519, next_seed)
        verkey = bytes_to_b58(keypair.get_public_bytes())
        try:
            await self._session.handle.insert_key(
                verkey,
                keypair,
            )
        except AskarError as err:
            if err.code == AskarErrorCode.DUPLICATE:
                pass
            else:
                raise WalletError(
                    "Error when creating new keypair for local DID"
                ) from err

        try:
            item = await self._session.handle.fetch(CATEGORY_DID, did, for_update=True)
            if not item:
                raise WalletNotFoundError("Unknown DID: {}".format(did)) from None
            entry_val = item.value_json
            metadata = entry_val.get("metadata", {})
            metadata["next_verkey"] = verkey
            entry_val["metadata"] = metadata
            await self._session.handle.replace(
                CATEGORY_DID, did, value_json=entry_val, tags=item.tags
            )
        except AskarError as err:
            raise WalletError("Error updating DID metadata") from err

        return verkey

    async def rotate_did_keypair_apply(self, did: str) -> DIDInfo:
        """Apply temporary keypair as main for DID that wallet owns.

        Args:
            did: signing DID

        Returns:
            DIDInfo with new verification key and metadata for DID

        """
        try:
            item = await self._session.handle.fetch(CATEGORY_DID, did, for_update=True)
            if not item:
                raise WalletNotFoundError("Unknown DID: {}".format(did)) from None
            entry_val = item.value_json
            metadata = entry_val.get("metadata", {})
            next_verkey = metadata.get("next_verkey")
            if not next_verkey:
                raise WalletError("Cannot rotate DID key: no next key established")
            del metadata["next_verkey"]
            entry_val["verkey"] = next_verkey
            item.tags["verkey"] = next_verkey
            await self._session.handle.replace(
                CATEGORY_DID, did, value_json=entry_val, tags=item.tags
            )
        except AskarError as err:
            raise WalletError("Error updating DID metadata") from err

    async def sign_message(
        self, message: Union[List[bytes], bytes], from_verkey: str
    ) -> bytes:
        """Sign message(s) using the private key associated with a given verkey.

        Args:
            message: The message(s) to sign
            from_verkey: Sign using the private key related to this verkey

        Returns:
            A signature

        Raises:
            WalletError: If the message is not provided
            WalletError: If the verkey is not provided
            WalletError: If another backend error occurs

        """
        if not message:
            raise WalletError("Message not provided")
        if not from_verkey:
            raise WalletError("Verkey not provided")
        try:
            keypair = await self._session.handle.fetch_key(from_verkey)
            if not keypair:
                raise WalletNotFoundError("Missing key for sign operation")
            key = keypair.key
            if key.algorithm == KeyAlg.BLS12_381_G2:
                # for now - must extract the key and use sign_message
                return sign_message(
                    message=message,
                    secret=key.get_secret_bytes(),
                    key_type=BLS12381G2,
                )

            else:
                return key.sign_message(message)
        except AskarError as err:
            raise WalletError("Exception when signing message") from err

    async def verify_message(
        self,
        message: Union[List[bytes], bytes],
        signature: bytes,
        from_verkey: str,
        key_type: KeyType,
    ) -> bool:
        """Verify a signature against the public key of the signer.

        Args:
            message: The message to verify
            signature: The signature to verify
            from_verkey: Verkey to use in verification
            key_type: The key type to derive the signature verification algorithm from

        Returns:
            True if verified, else False

        Raises:
            WalletError: If the verkey is not provided
            WalletError: If the signature is not provided
            WalletError: If the message is not provided
            WalletError: If another backend error occurs

        """
        if not from_verkey:
            raise WalletError("Verkey not provided")
        if not signature:
            raise WalletError("Signature not provided")
        if not message:
            raise WalletError("Message not provided")

        verkey = b58_to_bytes(from_verkey)

        if key_type == ED25519:
            try:
                pk = Key.from_public_bytes(KeyAlg.ED25519, verkey)
                return pk.verify_signature(message, signature)
            except AskarError as err:
                raise WalletError("Exception when verifying message signature") from err
        elif key_type == P256:
            try:
                pk = Key.from_public_bytes(KeyAlg.P256, verkey)
                return pk.verify_signature(message, signature)
            except AskarError as err:
                raise WalletError("Exception when verifying message signature") from err

        # other key types are currently verified outside of Askar
        return verify_signed_message(
            message=message,
            signature=signature,
            verkey=verkey,
            key_type=key_type,
        )

    async def pack_message(
        self, message: str, to_verkeys: Sequence[str], from_verkey: Optional[str] = None
    ) -> bytes:
        """Pack a message for one or more recipients.

        Args:
            message: The message to pack
            to_verkeys: List of verkeys for which to pack
            from_verkey: Sender verkey from which to pack

        Returns:
            The resulting packed message bytes

        Raises:
            WalletError: If no message is provided
            WalletError: If another backend error occurs

        """
        if message is None:
            raise WalletError("Message not provided")
        try:
            if from_verkey:
                from_key_entry = await self._session.handle.fetch_key(from_verkey)
                if not from_key_entry:
                    raise WalletNotFoundError("Missing key for pack operation")
                from_key = from_key_entry.key
            else:
                from_key = None
            return await asyncio.get_event_loop().run_in_executor(
                None, pack_message, to_verkeys, from_key, message
            )
        except AskarError as err:
            raise WalletError("Exception when packing message") from err

    async def unpack_message(self, enc_message: bytes) -> Tuple[str, str, str]:
        """Unpack a message.

        Args:
            enc_message: The packed message bytes

        Returns:
            A tuple: (message, from_verkey, to_verkey)

        Raises:
            WalletError: If the message is not provided
            WalletError: If another backend error occurs

        """
        if not enc_message:
            raise WalletError("Message not provided")
        try:
            (
                unpacked_json,
                recipient,
                sender,
            ) = await unpack_message(self._session.handle, enc_message)
        except AskarError as err:
            raise WalletError("Exception when unpacking message") from err
        return unpacked_json.decode("utf-8"), sender, recipient

    def _load_did_entry(self, entry: Entry) -> DIDInfo:
        """Convert a DID record into the expected DIDInfo format."""
        did_info = entry.value_json
        did_methods: DIDMethods = self._session.inject(DIDMethods)
        key_types: KeyTypes = self._session.inject(KeyTypes)
        return DIDInfo(
            did=did_info["did"],
            verkey=did_info["verkey"],
            metadata=did_info.get("metadata"),
            method=did_methods.from_method(did_info.get("method", "sov")) or SOV,
            key_type=key_types.from_key_type(did_info.get("verkey_type", "ed25519"))
            or ED25519,
        )


def _create_keypair(key_type: KeyType, seed: Union[str, bytes, None] = None) -> Key:
    """Instantiate a new keypair with an optional seed value."""
    if key_type == ED25519:
        alg = KeyAlg.ED25519
        method = None
    # elif key_type == BLS12381G1:
    #     alg = KeyAlg.BLS12_381_G1
    elif key_type == X25519:
        alg = KeyAlg.X25519
        method = None
    elif key_type == P256:
        alg = KeyAlg.P256
    elif key_type == BLS12381G2:
        alg = KeyAlg.BLS12_381_G2
        method = SeedMethod.BlsKeyGen
    # elif key_type == BLS12381G1G2:
    #     alg = KeyAlg.BLS12_381_G1G2
    else:
        raise WalletError(f"Unsupported key algorithm: {key_type}")
    if seed:
        try:
            if key_type in (ED25519, P256):
                # not a seed - it is the secret key
                seed = validate_seed(seed)
                return Key.from_secret_bytes(alg, seed)
            else:
                return Key.from_seed(alg, seed, method=method)
        except AskarError as err:
            if err.code == AskarErrorCode.INPUT:
                raise WalletError("Invalid seed for key generation") from err
            else:
                LOGGER.error(f"Unhandled Askar error code: {err.code}")
                raise
    else:
        return Key.generate(alg)



DIF_PRESENTATION_SUBMISSION_V1 = {
    "@context": {
        "@version": 1.1,
        "PresentationSubmission": {
            "@id": "https://identity.foundation/presentation-exchange/#presentation-submission",
            "@context": {
                "@version": 1.1,
                "presentation_submission": {
                    "@id": "https://identity.foundation/presentation-exchange/#presentation-submission",
                    "@type": "@json",
                },
            },
        },
    }
}



import pytest
import pytest_asyncio

from ......core.protocol_registry import ProtocolRegistry
from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......tests import mock
from ......utils.testing import create_test_profile
from ...handlers.query_handler import QueryHandler
from ...messages.disclose import Disclose
from ...messages.query import Query

TEST_MESSAGE_FAMILY = "doc/proto/1.0"
TEST_MESSAGE_TYPE = TEST_MESSAGE_FAMILY + "/message"


@pytest_asyncio.fixture
async def request_context():
    ctx = RequestContext.test_context(await create_test_profile())
    registry = ProtocolRegistry()
    registry.register_message_types({TEST_MESSAGE_TYPE: object()})
    profile = ctx.profile
    profile.context.injector.bind_instance(ProtocolRegistry, registry)
    yield ctx


class TestQueryHandler:
    @pytest.mark.asyncio
    async def test_query_all(self, request_context):
        query_msg = Query(query="*")
        query_msg.assign_thread_id("test123")
        request_context.message = query_msg
        request_context.connection_ready = True
        handler = QueryHandler()
        responder = MockResponder()
        await handler.handle(request_context, responder)
        messages = responder.messages
        assert len(messages) == 1
        result, target = messages[0]
        assert isinstance(result, Disclose) and result.protocols
        assert result.protocols[0]["pid"] == TEST_MESSAGE_FAMILY
        assert not target

    @pytest.mark.asyncio
    async def test_query_all_disclose_list_settings(self, request_context):
        profile = request_context.profile
        registry = profile.inject(ProtocolRegistry)
        registry.register_message_types({"doc/proto-b/1.0/message": object()})
        profile.context.injector.bind_instance(ProtocolRegistry, registry)
        profile.settings["disclose_protocol_list"] = [TEST_MESSAGE_FAMILY]
        query_msg = Query(query="*")
        query_msg.assign_thread_id("test123")
        request_context.message = query_msg
        request_context.connection_ready = True
        handler = QueryHandler()
        responder = MockResponder()
        await handler.handle(request_context, responder)
        messages = responder.messages
        assert len(messages) == 1
        result, target = messages[0]
        assert isinstance(result, Disclose) and result.protocols
        assert result.protocols[0]["pid"] == TEST_MESSAGE_FAMILY
        assert not target

    @pytest.mark.asyncio
    async def test_receive_query_process_disclosed(self, request_context):
        query_msg = Query(query="*")
        query_msg.assign_thread_id("test123")
        request_context.message = query_msg
        request_context.connection_ready = True
        handler = QueryHandler()
        responder = MockResponder()
        with (
            mock.patch.object(
                ProtocolRegistry, "protocols_matching_query", mock.MagicMock()
            ),
            mock.patch.object(
                ProtocolRegistry, "prepare_disclosed", mock.CoroutineMock()
            ) as mock_prepare_disclosed,
        ):
            mock_prepare_disclosed.return_value = [
                {"test": "test"},
                {
                    "pid": "https://didcomm.org/action-menu/1.0",
                    "roles": ["provider"],
                },
            ]
            await handler.handle(request_context, responder)
            messages = responder.messages
            assert len(messages) == 1
            result, target = messages[0]
            assert isinstance(result, Disclose) and result.protocols
            assert len(result.protocols) == 1
            assert result.protocols[0]["pid"] == "https://didcomm.org/action-menu/1.0"
            assert result.protocols[0]["roles"] == ["provider"]
            assert not target



"""Test keylist update response message."""

from unittest import TestCase

from ...message_types import KEYLIST_UPDATE_RESPONSE
from ..inner.keylist_updated import KeylistUpdated
from ..keylist_update_response import KeylistUpdateResponse, KeylistUpdateResponseSchema
from . import MessageTest


class TestKeylistUpdateResponse(MessageTest, TestCase):
    """Test keylist update response message."""

    TYPE = KEYLIST_UPDATE_RESPONSE
    CLASS = KeylistUpdateResponse
    SCHEMA = KeylistUpdateResponseSchema
    VALUES = {
        "updated": [
            KeylistUpdated(
                recipient_key="did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                action="added",
                result="success",
            )
        ]
    }



"""Transaction Job to send handler."""

from .....messaging.base_handler import (
    BaseHandler,
    BaseResponder,
    HandlerException,
    RequestContext,
)
from ..manager import TransactionManager, TransactionManagerError
from ..messages.transaction_job_to_send import TransactionJobToSend


class TransactionJobToSendHandler(BaseHandler):
    """Handler class for sending transaction jobs."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle transaction jobs.

        Args:
            context: Request context
            responder: Responder callback
        """

        self._logger.debug("TransactionJobToSendHandler called with context %s", context)
        assert isinstance(context.message, TransactionJobToSend)

        if not context.connection_ready:
            raise HandlerException("No connection established")
        assert context.connection_record

        mgr = TransactionManager(context.profile)
        try:
            await mgr.set_transaction_their_job(
                context.message, context.connection_record
            )
        except TransactionManagerError:
            self._logger.exception("Error receiving transaction jobs")



VACCINATION_V1 = {
    "@context": {
        "@version": 1.1,
        "@protected": True,
        "id": "@id",
        "type": "@type",
        "description": "http://schema.org/description",
        "identifier": "http://schema.org/identifier",
        "name": "http://schema.org/name",
        "image": "http://schema.org/image",
        "VaccinationCertificate": {
            "@id": "https://w3id.org/vaccination#VaccinationCertificate",
            "@context": {
                "@version": 1.1,
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "description": "http://schema.org/description",
                "identifier": "http://schema.org/identifier",
                "name": "http://schema.org/name",
                "image": "http://schema.org/image",
            },
        },
        "VaccinationEvent": {
            "@id": "https://w3id.org/vaccination#VaccinationEvent",
            "@context": {
                "@version": 1.1,
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "administeringCentre": "https://w3id.org/vaccination#administeringCentre",
                "batchNumber": "https://w3id.org/vaccination#batchNumber",
                "countryOfVaccination": "https://w3id.org/vaccination#countryOfVaccination",
                "dateOfVaccination": {
                    "@id": "https://w3id.org/vaccination#dateOfVaccination",
                    "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                },
                "healthProfessional": "https://w3id.org/vaccination#healthProfessional",
                "nextVaccinationDate": {
                    "@id": "https://w3id.org/vaccination#nextVaccinationDate",
                    "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                },
                "order": "https://w3id.org/vaccination#order",
                "recipient": {
                    "@id": "https://w3id.org/vaccination#recipient",
                    "@type": "https://w3id.org/vaccination#VaccineRecipient",
                },
                "vaccine": {
                    "@id": "https://w3id.org/vaccination#VaccineEventVaccine",
                    "@type": "https://w3id.org/vaccination#Vaccine",
                },
            },
        },
        "VaccineRecipient": {
            "@id": "https://w3id.org/vaccination#VaccineRecipient",
            "@context": {
                "@version": 1.1,
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "birthDate": {
                    "@id": "http://schema.org/birthDate",
                    "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                },
                "familyName": "http://schema.org/familyName",
                "gender": "http://schema.org/gender",
                "givenName": "http://schema.org/givenName",
            },
        },
        "Vaccine": {
            "@id": "https://w3id.org/vaccination#Vaccine",
            "@context": {
                "@version": 1.1,
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "atcCode": "https://w3id.org/vaccination#atc-code",
                "disease": "https://w3id.org/vaccination#disease",
                "event": {
                    "@id": "https://w3id.org/vaccination#VaccineRecipientVaccineEvent",
                    "@type": "https://w3id.org/vaccination#VaccineEvent",
                },
                "marketingAuthorizationHolder": "https://w3id.org/vaccination#marketingAuthorizationHolder",
                "medicinalProductName": "https://w3id.org/vaccination#medicinalProductName",
            },
        },
    }
}



"""Represents a transaction resend message."""

from typing import Optional

from marshmallow import EXCLUDE, fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from .....messaging.valid import UUID4_EXAMPLE
from ..message_types import PROTOCOL_PACKAGE, TRANSACTION_RESEND

HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers.transaction_resend_handler.TransactionResendHandler"
)


class TransactionResend(AgentMessage):
    """Class representing a transaction resend message."""

    class Meta:
        """Metadata for a transaction resend message."""

        handler_class = HANDLER_CLASS
        message_type = TRANSACTION_RESEND
        schema_class = "TransactionResendSchema"

    def __init__(
        self,
        *,
        state: Optional[str] = None,
        thread_id: Optional[str] = None,
        **kwargs,
    ):
        """Initialize a transaction resend object.

        Args:
            state: State of the transaction record
            thread_id: Thread id of transaction record
            kwargs: Additional keyword arguments for the message
        """
        super().__init__(**kwargs)

        self.state = state
        self.thread_id = thread_id


class TransactionResendSchema(AgentMessageSchema):
    """Transaction resend schema class."""

    class Meta:
        """Transaction resend schema metadata."""

        model_class = TransactionResend
        unknown = EXCLUDE

    state = fields.Str(
        required=False,
        metadata={
            "description": "The State of the transaction Record",
            "example": "resend",
        },
    )
    thread_id = fields.Str(required=False, metadata={"example": UUID4_EXAMPLE})



from unittest import TestCase

from ...wallet.key_type import BLS12381G1, BLS12381G1G2, BLS12381G2
from ...wallet.util import b58_to_bytes
from ..did_key import DID_KEY_RESOLVERS, DIDKey
from .test_dids import (
    DID_BLS12381G1G2_z5TcESXuYUE9aZWYwSdrUEGK1HNQFHyTt4aVpaCTVZcDXQmUheFwfNZmRksaAbBneNm5KyE52SdJeRCN1g6PJmF31GsHWwFiqUDujvasK3wTiDr3vvkYwEJHt7H5RGEKYEp1ErtQtcEBgsgY2DA9JZkHj1J9HZ8MRDTguAhoFtR4aTBQhgnkP4SwVbxDYMEZoF2TMYn3s,
)

TEST_BLS12381G1G2_BASE58_KEY = "AQ4MiG1JKHmM5N4CgkF9uQ484PHN7gXB3ctF4ayL8hT6FdD6rcfFS3ZnMNntYsyJBckfNPf3HL8VU8jzgyT3qX88Yg3TeF2NkG2aZnJDNnXH1jkJStWMxjLw22LdphqAj1rSorsDhHjE8Rtz61bD6FP9aPokQUDVpZ4zXqsXVcxJ7YEc66TTLTTPwQPS7uNM4u2Fs"
TEST_BLS12381G1G2_FINGERPRINT = "z5TcESXuYUE9aZWYwSdrUEGK1HNQFHyTt4aVpaCTVZcDXQmUheFwfNZmRksaAbBneNm5KyE52SdJeRCN1g6PJmF31GsHWwFiqUDujvasK3wTiDr3vvkYwEJHt7H5RGEKYEp1ErtQtcEBgsgY2DA9JZkHj1J9HZ8MRDTguAhoFtR4aTBQhgnkP4SwVbxDYMEZoF2TMYn3s"
TEST_BLS12381G1G2_DID = f"did:key:{TEST_BLS12381G1G2_FINGERPRINT}"

TEST_BLS12381G1_BASE58_KEY = (
    "7BVES4h78wzabPAfMhchXyH5d8EX78S5TtzePH2YkftWcE6by9yj3NTAv9nsyCeYch"
)
TEST_BLS12381G1_FINGERPRINT = (
    "z3tEG5qmJZX29jJSX5kyhDR5YJNnefJFdwTxRqk6zbEPv4Pf2xF12BpmXv9NExxSRFGfxd"
)
TEST_BLS12381G1_DID = f"did:key:{TEST_BLS12381G1_FINGERPRINT}"

TEST_BLS12381G2_BASE58_KEY = "26d2BdqELsXg7ZHCWKL2D5Y2S7mYrpkdhJemSEEvokd4qy4TULJeeU44hYPGKo4x4DbBp5ARzkv1D6xuB3bmhpdpKAXuXtode67wzh9PCtW8kTqQhH19VSiFZkLNkhe9rtf3"
TEST_BLS12381G2_FINGERPRINT = "zUC7LTa4hWtaE9YKyDsMVGiRNqPMN3s4rjBdB3MFi6PcVWReNfR72y3oGW2NhNcaKNVhMobh7aHp8oZB3qdJCs7RebM2xsodrSm8MmePbN25NTGcpjkJMwKbcWfYDX7eHCJjPGM"
TEST_BLS12381G2_DID = f"did:key:{TEST_BLS12381G2_FINGERPRINT}"
TEST_BLS12381G1G2_PREFIX_BYTES = b"".join(
    [b"\xee\x01", b58_to_bytes(TEST_BLS12381G1G2_BASE58_KEY)]
)


# The tests here are a bit quirky because g1g2 is a concatenation of g1 and g2 public key bytes
# but it works with the already existing did key implementation.
class TestDIDKey(TestCase):
    def test_bls12381g1g2_from_public_key(self):
        key_bytes = b58_to_bytes(TEST_BLS12381G1G2_BASE58_KEY)
        did_key = DIDKey.from_public_key(key_bytes, BLS12381G1G2)

        assert did_key.did == TEST_BLS12381G1G2_DID

    def test_bls12381g1g2_from_public_key_b58(self):
        did_key = DIDKey.from_public_key_b58(TEST_BLS12381G1G2_BASE58_KEY, BLS12381G1G2)

        assert did_key.did == TEST_BLS12381G1G2_DID

    def test_bls12381g1g2_from_fingerprint(self):
        did_key = DIDKey.from_fingerprint(TEST_BLS12381G1G2_FINGERPRINT)

        assert did_key.did == TEST_BLS12381G1G2_DID
        assert did_key.public_key_b58 == TEST_BLS12381G1G2_BASE58_KEY

    def test_bls12381g1g2_from_did(self):
        did_key = DIDKey.from_did(TEST_BLS12381G1G2_DID)

        assert did_key.public_key_b58 == TEST_BLS12381G1G2_BASE58_KEY

    def test_bls12381g1g2_properties(self):
        did_key = DIDKey.from_did(TEST_BLS12381G1G2_DID)

        assert did_key.fingerprint == TEST_BLS12381G1G2_FINGERPRINT
        assert did_key.did == TEST_BLS12381G1G2_DID
        assert did_key.public_key_b58 == TEST_BLS12381G1G2_BASE58_KEY
        assert did_key.public_key == b58_to_bytes(TEST_BLS12381G1G2_BASE58_KEY)
        assert did_key.key_type == BLS12381G1G2
        assert did_key.prefixed_public_key == TEST_BLS12381G1G2_PREFIX_BYTES

    def test_bls12381g1g2_diddoc(self):
        did_key = DIDKey.from_did(TEST_BLS12381G1G2_DID)

        resolver = DID_KEY_RESOLVERS[BLS12381G1G2]

        assert resolver(did_key) == did_key.did_doc

    def test_bls12381g1g2_resolver(self):
        did_key = DIDKey.from_did(
            "did:key:z5TcESXuYUE9aZWYwSdrUEGK1HNQFHyTt4aVpaCTVZcDXQmUheFwfNZmRksaAbBneNm5KyE52SdJeRCN1g6PJmF31GsHWwFiqUDujvasK3wTiDr3vvkYwEJHt7H5RGEKYEp1ErtQtcEBgsgY2DA9JZkHj1J9HZ8MRDTguAhoFtR4aTBQhgnkP4SwVbxDYMEZoF2TMYn3s"
        )
        resolver = DID_KEY_RESOLVERS[BLS12381G1G2]
        did_doc = resolver(did_key)

        assert (
            did_doc
            == DID_BLS12381G1G2_z5TcESXuYUE9aZWYwSdrUEGK1HNQFHyTt4aVpaCTVZcDXQmUheFwfNZmRksaAbBneNm5KyE52SdJeRCN1g6PJmF31GsHWwFiqUDujvasK3wTiDr3vvkYwEJHt7H5RGEKYEp1ErtQtcEBgsgY2DA9JZkHj1J9HZ8MRDTguAhoFtR4aTBQhgnkP4SwVbxDYMEZoF2TMYn3s
        )

    def test_bls12381g1g1_to_g1(self):
        g1g2_did = DIDKey.from_did(TEST_BLS12381G1G2_DID)

        # TODO: add easier method to go form g1 <- g1g2 -> g2
        # First 48 bytes is g1 key
        g1_public_key = g1g2_did.public_key[:48]
        g1_did = DIDKey.from_public_key(g1_public_key, BLS12381G1)

        assert g1_did.fingerprint == TEST_BLS12381G1_FINGERPRINT
        assert g1_did.did == TEST_BLS12381G1_DID
        assert g1_did.public_key_b58 == TEST_BLS12381G1_BASE58_KEY
        assert g1_did.public_key == b58_to_bytes(TEST_BLS12381G1_BASE58_KEY)
        assert g1_did.key_type == BLS12381G1

    def test_bls12381g1g1_to_g2(self):
        g1g2_did = DIDKey.from_did(TEST_BLS12381G1G2_DID)

        # TODO: add easier method to go form g1 <- g1g2 -> g2
        # From 48 bytes is g2 key
        g2_public_key = g1g2_did.public_key[48:]
        g2_did = DIDKey.from_public_key(g2_public_key, BLS12381G2)

        assert g2_did.fingerprint == TEST_BLS12381G2_FINGERPRINT
        assert g2_did.did == TEST_BLS12381G2_DID
        assert g2_did.public_key_b58 == TEST_BLS12381G2_BASE58_KEY
        assert g2_did.public_key == b58_to_bytes(TEST_BLS12381G2_BASE58_KEY)
        assert g2_did.key_type == BLS12381G2



"""Revocation utilities."""

import re
from typing import Optional, Sequence

from ..core.profile import Profile

REVOCATION_EVENT_PREFIX = "acapy::REVOCATION::"
EVENT_LISTENER_PATTERN = re.compile(f"^{REVOCATION_EVENT_PREFIX}(.*)?$")
REVOCATION_REG_INIT_EVENT = "REGISTRY_INIT"
REVOCATION_REG_ENDORSED_EVENT = "REGISTRY_ENDORSED"
REVOCATION_ENTRY_ENDORSED_EVENT = "ENTRY_ENDORSED"
REVOCATION_ENTRY_EVENT = "SEND_ENTRY"
REVOCATION_PUBLISHED_EVENT = "published"
REVOCATION_CLEAR_PENDING_EVENT = "clear-pending"


async def notify_revocation_reg_init_event(
    profile: Profile,
    issuer_rev_id: str,
    create_pending_rev_reg: bool = False,
    endorser_connection_id: Optional[str] = None,
):
    """Send notification for a revocation registry init event."""
    meta_data = {
        "context": {
            "issuer_rev_id": issuer_rev_id,
        },
        "processing": {"create_pending_rev_reg": create_pending_rev_reg},
    }
    if endorser_connection_id:
        meta_data["endorser"] = {"connection_id": endorser_connection_id}
    topic = f"{REVOCATION_EVENT_PREFIX}{REVOCATION_REG_INIT_EVENT}::{issuer_rev_id}"
    await profile.notify(topic, meta_data)


async def notify_revocation_entry_event(
    profile: Profile, issuer_rev_id: str, meta_data: dict
):
    """Send notification for a revocation registry entry event."""
    topic = f"{REVOCATION_EVENT_PREFIX}{REVOCATION_ENTRY_EVENT}::{issuer_rev_id}"
    await profile.notify(topic, meta_data)


async def notify_revocation_reg_endorsed_event(
    profile: Profile, rev_reg_id: str, meta_data: dict
):
    """Send notification for a revocation registry endorsement event."""
    topic = f"{REVOCATION_EVENT_PREFIX}{REVOCATION_REG_ENDORSED_EVENT}::{rev_reg_id}"
    await profile.notify(topic, meta_data)


async def notify_revocation_entry_endorsed_event(
    profile: Profile, rev_reg_id: str, meta_data: dict, revoked: list
):
    """Send notification for a revocation registry entry endorsement event."""
    topic = f"{REVOCATION_EVENT_PREFIX}{REVOCATION_ENTRY_ENDORSED_EVENT}::{rev_reg_id}"
    await profile.notify(topic, meta_data)
    await notify_revocation_published_event(profile, rev_reg_id, revoked)


async def notify_revocation_published_event(
    profile: Profile,
    rev_reg_id: str,
    crids: Sequence[str],
):
    """Send notification of credential revoked as issuer."""
    topic = f"{REVOCATION_EVENT_PREFIX}{REVOCATION_PUBLISHED_EVENT}::{rev_reg_id}"
    await profile.notify(topic, {"rev_reg_id": rev_reg_id, "crids": crids})


async def notify_pending_cleared_event(
    profile: Profile,
    rev_reg_id: str,
):
    """Send notification of credential revoked as issuer."""
    topic = f"{REVOCATION_EVENT_PREFIX}{REVOCATION_CLEAR_PENDING_EVENT}::{rev_reg_id}"
    await profile.notify(topic, {"rev_reg_id": rev_reg_id})


async def notify_rev_reg_entry_txn_failed(
    profile: Profile,
    msg: str,
):
    """Send notification that a revocation registry entry transaction failed."""
    topic = f"{REVOCATION_EVENT_PREFIX}REV_REG_ENTRY_TXN_FAILED"
    await profile.notify(topic, {"msg": msg})



from unittest import TestCase, mock

from .....didcomm_prefix import DIDCommPrefix
from ...message_types import DISCLOSE, PROTOCOL_PACKAGE
from ..disclose import Disclose


class TestDisclose(TestCase):
    test_protocols = [
        {
            "pid": DIDCommPrefix.qualify_current("basicmessage/1.0/message"),
            "roles": [],
        }
    ]

    def test_init(self):
        disclose = Disclose(protocols=self.test_protocols)
        assert disclose.protocols == self.test_protocols

    def test_type(self):
        disclose = Disclose(protocols=self.test_protocols)
        assert disclose._type == DIDCommPrefix.qualify_current(DISCLOSE)

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.disclose.DiscloseSchema.load")
    def test_deserialize(self, mock_disclose_schema_load):
        obj = {"obj": "obj"}

        disclose = Disclose.deserialize(obj)
        mock_disclose_schema_load.assert_called_once_with(obj)

        assert disclose is mock_disclose_schema_load.return_value

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.disclose.DiscloseSchema.dump")
    def test_serialize(self, mock_disclose_schema_dump):
        disclose = Disclose(protocols=self.test_protocols)

        disclose_dict = disclose.serialize()
        mock_disclose_schema_dump.assert_called_once_with(disclose)

        assert disclose_dict is mock_disclose_schema_dump.return_value


class TestDiscloseSchema(TestCase):
    disclose = Disclose(protocols=[])

    def test_make_model(self):
        data = self.disclose.serialize()
        model_instance = Disclose.deserialize(data)
        assert isinstance(model_instance, Disclose)
